# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Integration authorization tests
#
# pylint: disable=consider-using-enumerate
# pylint: disable=too-many-locals
# pylint: disable=too-many-nested-blocks

import random
import time
import unittest

from okera.concurrency import OkeraWorkerException
from okera._thrift_api import TAuthorizeQueryParams
from okera._thrift_api import TRecordServiceException
from okera._thrift_api import TTypeId
from okera.tests import pycerebro_test_common as common

SKIP_LEVELS = ["smoke", "dev", "all"]

EPOCH0_DATE = '1970-01-01'
EPOCH0_TS = '1970-01-01 00:00:00.000'

TEST_DB = 'auth_test_db'
TEST_ROLE = 'auth_test_role'
TEST_USER = 'auth_test_user'
ABAC_TEST_ROLE = 'abac_test_role'
ABAC_TEST_USER = 'abac_test_user'
ABAC_TRANSFORM_ROLE = 'abac_transform_role'
ABAC_TRANSFORM_USER = 'abac_transform_user'
ABAC_TRANSFORM_ROLE2 = 'abac_transform_role2'
ABAC_TRANSFORM_USER2 = 'abac_transform_user2'

# Environment configs to control test running
MINIMAL_NUM_COLS = common.get_env_var('AUTH_TEST_MINIMAL_NUM_COLS', int, 1)
if common.get_env_var('AUTH_TEST_RANDOM_SEED', bool, False):
    print("Initializing to random seed (based on current time).")
    random.seed(None)
else:
    random.seed(1)
SKIP_SCANS = common.get_bool_env_var('AUTH_TEST_SKIP_SCANS', False)
AUTH_QUERY = SKIP_SCANS = common.get_bool_env_var('AUTH_TEST_AUTH_QUERY', False)

class AuthorizationTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """ Initializes one time state that is shared across test cases. This is used
            to speed up the tests. State that can be shared across (but still stable)
            should be here instead of __cleanup()."""
        super(AuthorizationTest, cls).setUpClass()
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.delete_attribute('abac_auth_test', 'v1')
            conn.create_attribute('abac_auth_test', 'v1')

    # Runs a SQL statement and log it
    @staticmethod
    def __ddl(conn, sql):
        print(sql)
        conn.execute_ddl(sql)

    def __cleanup(self, conn):
        """ Cleanups all the test state used in this test to "reset" the catalog.
            dbs can be specified to do the initialize over multiple databases.
            This can be used for tests that use multiple dbs (but makes the test
            take longer). By default, only load TEST_DB.
        """
        self.__ddl(conn, "DROP ROLE IF EXISTS %s" % TEST_ROLE)
        self.__ddl(conn, "CREATE ROLE %s" % TEST_ROLE)
        self.__ddl(conn, "GRANT ROLE %s to GROUP %s" % (TEST_ROLE, TEST_USER))
        self.__ddl(conn, "DROP ROLE IF EXISTS %s" % ABAC_TEST_ROLE)
        self.__ddl(conn, "CREATE ROLE %s" % ABAC_TEST_ROLE)
        self.__ddl(conn, "GRANT ROLE %s to GROUP %s" % (ABAC_TEST_ROLE, ABAC_TEST_USER))
        self.__ddl(conn, "DROP ROLE IF EXISTS %s" % ABAC_TRANSFORM_ROLE)
        self.__ddl(conn, "CREATE ROLE %s" % ABAC_TRANSFORM_ROLE)
        self.__ddl(conn, "GRANT ROLE %s to GROUP %s" % \
            (ABAC_TRANSFORM_ROLE, ABAC_TRANSFORM_USER))
        self.__ddl(conn, "DROP ROLE IF EXISTS %s" % ABAC_TRANSFORM_ROLE2)
        self.__ddl(conn, "CREATE ROLE %s" % ABAC_TRANSFORM_ROLE2)
        self.__ddl(conn, "GRANT ROLE %s to GROUP %s" % \
            (ABAC_TRANSFORM_ROLE2, ABAC_TRANSFORM_USER2))

    def __convert_col(self, cols, idx):
        name = '`' + cols[idx].name + '`'
        type = cols[idx].type
        children = []
        if cols[idx].type.num_children:
            idx += 1
            for _ in range(0, cols[idx - 1].type.num_children):
                idx, child_name, child = self.__convert_col(cols, idx)
                children.append((child_name, child))
        else:
            idx += 1
        return idx, name, (type, children)

    def __top_level_columns(self, schema):
        cols = schema.nested_cols
        if not cols:
            cols = schema.cols
        names = []
        types = {}
        idx = 0
        while idx < len(cols):
            idx, name, col = self.__convert_col(cols, idx)
            names.append(name)
            types[name] = col
        return names, types

    @staticmethod
    def __collect_grant_objects(conn):
        grants = conn.execute_ddl('SHOW GRANT ROLE %s' % (TEST_ROLE))
        result = []
        for grant in grants:
            path = ''
            if grant[1]:
                path = grant[1]
                if grant[2]:
                    path += '.' + grant[2]
                    if grant[3]:
                        path += '.' + grant[3]
                else:
                    path += '.*'
            else:
                path = '*'
            result.append(path)
        return result

    # Tests that revokes to a db does not cascade to the table or columns
    def test_revoke_db_no_cascade(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self.__cleanup(conn)

            # Grant on db, table and column
            conn.execute_ddl(
                'GRANT SELECT ON DATABASE okera_sample TO ROLE %s' % (TEST_ROLE))
            conn.execute_ddl(
                'GRANT SELECT ON TABLE okera_sample.sample TO ROLE %s' % (TEST_ROLE))
            conn.execute_ddl(
                'GRANT SELECT(record) ON TABLE okera_sample.sample TO ROLE %s' %\
                (TEST_ROLE))

            objs = self.__collect_grant_objects(conn)
            print(objs)
            self.assertTrue('okera_sample.*' in objs)
            self.assertTrue('okera_sample.sample' in objs)
            self.assertTrue('okera_sample.sample.record' in objs)
            self.assertEqual(len(objs), 3)

            # Revoke on db, should not cascade, this cascade can take a while so sleep
            # first. In the test setup, the refresh is 5 seconds.
            conn.execute_ddl(
                'REVOKE SELECT ON DATABASE okera_sample FROM ROLE %s' % (TEST_ROLE))
            time.sleep(7)
            objs = self.__collect_grant_objects(conn)
            print(objs)
            self.assertTrue('okera_sample.sample' in objs)
            self.assertTrue('okera_sample.sample.record' in objs)
            self.assertEqual(len(objs), 2)

            # Revoke on table, should not cascade.
            conn.execute_ddl(
                'REVOKE SELECT ON TABLE okera_sample.sample FROM ROLE %s' % (TEST_ROLE))
            time.sleep(7)
            objs = self.__collect_grant_objects(conn)
            print(objs)
            self.assertTrue('okera_sample.sample.record' in objs)
            self.assertEqual(len(objs), 1)

            # Revoke on column.
            conn.execute_ddl(
                'REVOKE SELECT(record) ON TABLE okera_sample.sample FROM ROLE %s' %\
                (TEST_ROLE))
            objs = self.__collect_grant_objects(conn)
            self.assertEqual(len(objs), 0)

    @staticmethod
    def __is_complex_type(t):
        return t.type_id == TTypeId.RECORD \
            or t.type_id == TTypeId.ARRAY \
            or t.type_id == TTypeId.MAP

    # Returns true if this value maps to 0
    def __is_zero(self, v):
        if isinstance(v, dict):
            for _, val in v.items():
                if not self.__is_zero(val):
                    return False
            return True
        if v is not None and isinstance(v, (bytes, str)):
            v = v.strip()
        return v in [0, '', b''] or str(v) in (EPOCH0_DATE, EPOCH0_TS) or float(v) == 0

    # Returns the first primitive type for the schema at col/t. Returns None if
    # all leaf types are complex (e.g. array)
    # FIXME: disable_recursion returns None if it is a record type. This is used
    # to disable some tests due to existing bugs.
    def __first_primitive_type(self, col, t, disable_recursion=False):
        if t[0].type_id == TTypeId.ARRAY or t[0].type_id == TTypeId.MAP:
            return None
        if t[0].type_id != TTypeId.RECORD:
            return col
        if disable_recursion:
            return None

        for child in t[1]:
            child_name = self.__first_primitive_type(
                child[0], child[1], disable_recursion)
            if not child_name:
                continue
            return col + '.' + child_name
        return None

    # Returns the number of fully expanded columns for this type. i.e. flattened
    # structs
    def __expanded_result_cols(self, t):
        result = 0
        if t[0].type_id == TTypeId.RECORD:
            for child in t[1]:
                result += self.__expanded_result_cols(child[1])
            return result
        if t[0].type_id == TTypeId.ARRAY:
            return 1 + self.__expanded_result_cols(t[1][0][1])
        if t[0].type_id == TTypeId.MAP:
            return 1 + self.__expanded_result_cols(t[1][0][1]) +\
                self.__expanded_result_cols(t[1][1][1])
        return 1

    # Verify that this query is authorized. If analysis_error is true, it means this
    # query is known to be semantically invalid but is authorized. This happens when
    # we, for example, run transformations on complex types.
    def __test_valid_query(self, conn, user, query, num_result_cols, test_scan,
                           analysis_error=False, pandas_num_result_cols=None,
                           expect_no_results=False, expect_complex_types_error=False,
                           transformed_cols=None):
        print('Running valid query: ' + query)
        if analysis_error:
            # We expect an analysis error (not auth) as this uses functions on complex
            # types
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                conn.plan(query)
            self.assertTrue('does not support complex types' in str(ex_ctx.exception) or\
                'No matching function' in str(ex_ctx.exception) or\
                'this column is a complex type' in str(ex_ctx.exception))
        else:
            try:
                schema = conn.plan(query).schema
            except TRecordServiceException as ex:
                # Transformations on complex types will fail.
                if expect_complex_types_error and \
                      user in [ABAC_TRANSFORM_USER, ABAC_TRANSFORM_USER2]:
                    self.assertTrue('this column is a complex type' in str(ex.detail))
                    return
                raise

            result_cols, _ = self.__top_level_columns(schema)
            self.assertEqual(num_result_cols, len(result_cols))

            # Verify authorize query succeeds
            if AUTH_QUERY:
                authorize_query = TAuthorizeQueryParams()
                authorize_query.sql = query
                authorize_query_result = conn.service.client.AuthorizeQuery(
                    authorize_query)
                self.assertTrue(authorize_query_result is not None)

            if not test_scan or SKIP_SCANS:
                return

            # Read the data from json and pandas
            try:
                result = conn.scan_as_json(query, max_records=1, max_task_count=1,
                                           strings_as_utf8=False)
                if expect_no_results:
                    self.assertEqual(0, len(result), msg="Result should have been empty.")
                if result:
                    self.assertEqual(num_result_cols, len(result[0]))
                    if transformed_cols and 'count(' not in query:
                        for row in result:
                            for key, v in row.items():
                                key = key.split('.')[0]
                                if key in transformed_cols:
                                    self.assertTrue(self.__is_zero(v))
            except OkeraWorkerException as ex:
                self.assertTrue('Only arrays of string, int32' in str(ex.error))
            try:
                df = conn.scan_as_pandas(query, max_records=1, max_task_count=1)
                if expect_no_results:
                    self.assertEqual(0, len(df), msg="Result should have been empty.")
                self.assertTrue(
                    len(df.columns) in [num_result_cols, pandas_num_result_cols],
                    msg=df)
            except OkeraWorkerException as ex:
                self.assertTrue('Only arrays of string, int32' in str(ex.error))

    # Verifies this query fails with an auth error
    def __test_invalid_query(self, conn, query):
        print("Running invalid query: " + query)
        with self.assertRaises(TRecordServiceException) as ex_ctx:
            conn.plan(query)
        self.assertTrue(
            'not have privilege' in str(ex_ctx.exception) or\
            'column is a complex type' in str(ex_ctx.exception),
            msg=str(ex_ctx.exception))

    def __test_visible_cols(self, conn, db, ds, user, types, visible_cols,
                            transformed_cols, test_scan):
        all_cols = ', '.join(visible_cols)
        tbl = db + '.' + ds

        # Queries that return all columns
        all_queries = [
            'SELECT * FROM %s.%s' % (db, ds),
            'SELECT %s from %s.%s' % (all_cols, db, ds),
            'SELECT * FROM (SELECT %s from %s.%s)v' % (all_cols, db, ds),
            'SELECT %s FROM (SELECT %s from %s.%s)v' % (all_cols, all_cols, db, ds),
            'SELECT %s FROM ( WITH cte_tbl as ' \
            '(SELECT %s from %s.%s) SELECT * FROM cte_tbl) v' %
            (all_cols, all_cols, db, ds),
        ]

        contains_complex = False
        for col in visible_cols:
            contains_complex |= self.__is_complex_type(types[col][0])

        for query in all_queries:
            print('Running valid query: ' + query)
            try:
                schema = conn.plan(query).schema
                result_cols, _ = self.__top_level_columns(schema)
                self.assertEqual(len(visible_cols), len(result_cols))
            except TRecordServiceException as ex:
                # Transformations on complx types will fail.
                if contains_complex and \
                      user in [ABAC_TRANSFORM_USER, ABAC_TRANSFORM_USER2]:
                    self.assertTrue('this column is a complex type' in str(ex.detail))
                else:
                    raise

        first_col = visible_cols[0]
        first_col_expanded = self.__expanded_result_cols(types[first_col])
        last_col = visible_cols[-1]
        last_col_complex = self.__is_complex_type(types[last_col][0])
        last_col_expanded = self.__expanded_result_cols(types[last_col])

        # Run some specific queries
        self.__test_valid_query(
            conn, user, 'SELECT %s FROM %s.%s' % (last_col, db, ds), 1,
            test_scan, False, last_col_expanded,
            expect_complex_types_error=self.__is_complex_type(types[last_col][0]),
            transformed_cols=transformed_cols)
        self.__test_valid_query(
            conn, user, 'SELECT count(*) FROM %s.%s' % (db, ds), 1, test_scan,
            transformed_cols=transformed_cols)

        # Go into the struct type to find a primitive type to query
        last_col_primitive = self.__first_primitive_type(last_col, types[last_col])
        first_col_primitive = self.__first_primitive_type(first_col, types[first_col])
        if last_col_primitive:
            self.__test_valid_query(
                conn, user, 'SELECT count(%s) FROM %s.%s' % (last_col_primitive, db, ds),
                1, test_scan, transformed_cols=transformed_cols)
            self.__test_valid_query(
                conn, user, 'SELECT count(%s) FROM %s.%s WHERE %s IS NOT NULL' % \
                    (last_col_primitive, db, ds, last_col_primitive), 1,
                test_scan, transformed_cols=transformed_cols)
            self.__test_valid_query(
                conn, user, 'SELECT count(%s) as c1, count(%s) as c2 FROM %s.%s' % \
                    (last_col_primitive, last_col_primitive, db, ds), 2,
                test_scan, False, transformed_cols=transformed_cols)

            if first_col_primitive and len(visible_cols) > 1:
                self.__test_valid_query(
                    conn, user, 'SELECT count(%s) as c1, count(%s) as c2 FROM %s.%s' % \
                        (first_col_primitive, last_col_primitive, db, ds), 2,
                    test_scan, transformed_cols=transformed_cols)

        # FIXME: for fields in structs, fails with:
        # Illegal reference to non-materialized slot
        last_col_primitive = self.__first_primitive_type(last_col, types[last_col], True)
        first_col_complex = self.__is_complex_type(types[first_col][0])
        if last_col_primitive:
            # Join on accessible column
            self.__test_valid_query(
                conn, user,
                'select a.%s from %s a join %s b on a.%s = b.%s' %\
                    (last_col, tbl, tbl, last_col_primitive, last_col_primitive),
                1, test_scan, False, first_col_expanded,
                transformed_cols=transformed_cols)
            self.__test_valid_query(
                conn, user,
                'select a.%s from %s a join %s b on a.%s = b.%s' %\
                    (first_col, tbl, tbl, last_col_primitive, last_col_primitive),
                1, test_scan, False, first_col_expanded,
                expect_complex_types_error=first_col_complex,
                transformed_cols=transformed_cols)

        # Get a primitive column.
        # This should be collapsed into the last block once above bugs are fixed
        last_col_primitive = self.__first_primitive_type(last_col, types[last_col])
        if last_col_primitive:
            # Test some queries that should return no results due to the where clause
            empty_where = 'CAST(%s AS STRING) = "NOT A VALUE TEST"' % last_col_primitive
            empty_queries = [
                'SELECT %s FROM %s.%s WHERE %s' % (first_col, db, ds, empty_where),
                'SELECT %s FROM (SELECT %s FROM %s.%s WHERE %s)v' %\
                    (first_col, first_col, db, ds, empty_where),
                'SELECT * FROM (SELECT %s FROM %s.%s WHERE %s)v' %\
                    (first_col, db, ds, empty_where),
            ]
            for query in empty_queries:
                self.__test_valid_query(
                    conn, user, query, 1, test_scan, False, first_col_expanded, True,
                    expect_complex_types_error=first_col_complex,
                    transformed_cols=transformed_cols)

            empty_queries = [
                'SELECT %s FROM (SELECT * FROM %s.%s WHERE %s)v' %\
                    (first_col, db, ds, empty_where),
            ]
            for query in empty_queries:
                self.__test_valid_query(
                    conn, user, query, 1, test_scan, False, first_col_expanded, True,
                    expect_complex_types_error=contains_complex,
                    transformed_cols=transformed_cols)

        # Run some queries that return one column
        queries_one_col_result = [
            'SELECT count(%s) FROM %s.%s' % (last_col, db, ds),
            'SELECT %s FROM %s.%s WHERE %s IS NOT NULL' % (first_col, db, ds, last_col),
            'SELECT count(%s) FROM %s.%s WHERE %s IS NOT NULL' % \
                (last_col, db, ds, last_col),
        ]
        for query in queries_one_col_result:
            is_complex = self.__is_complex_type(types[first_col][0])
            is_complex |= self.__is_complex_type(types[last_col][0])
            self.__test_valid_query(
                conn, user, query, 1, test_scan, last_col_complex, first_col_expanded,
                expect_complex_types_error=is_complex,
                transformed_cols=transformed_cols)

    def __test_no_access(self, conn, db, ds):
        # Tests when the user has access to nothing in the dataset
        # when no columns eligible for select, a privilege exception is thrown.
        queries = [
            'SELECT * FROM %s.%s' % (db, ds),
            'SELECT count(*) FROM %s.%s' % (db, ds),
            'SELECT count(*) FROM %s.%s WHERE false' % (db, ds),
            'SELECT * FROM (SELECT * FROM %s.%s)v' % (db, ds),
            'SELECT count(*) FROM (SELECT * FROM %s.%s)v' % (db, ds),
            'SELECT count(*) FROM (SELECT * FROM %s.%s)v' % (db, ds),
            'SELECT * FROM ( WITH cte_cnt_tbl as ' \
            '(SELECT * from %s.%s) SELECT count(*) FROM cte_cnt_tbl) v' % (db, ds)
        ]
        for query in queries:
            self.__test_invalid_query(conn, query)

    # Authorization tests where the user has access to visible_cols but not
    # non_visible_col.
    def __test_partial_access(self, conn, db, ds, user, types, visible_cols,
                              non_visible_col):
        tbl = db + '.' + ds
        visible = 1
        visible_col = 1
        visible_col_primitive = None
        if visible_cols:
            visible = ', '.join(visible_cols)
            visible_col = visible_cols[0]
            visible_col_primitive = self.__first_primitive_type(
                visible_col, types[visible_col])
        non_visible_col_primitive = self.__first_primitive_type(
            non_visible_col, types[non_visible_col])

        contains_complex = False
        for col in visible_cols:
            contains_complex |= self.__is_complex_type(types[col][0])

        # Collect all the queries that should fail
        queries = [
            'select %s from %s' % (non_visible_col, tbl)
        ]

        # Queries with both visible and not, should fail
        queries += [
            'select %s from %s WHERE %s IS NULL' % (visible_col, tbl, non_visible_col),
            'select %s from %s WHERE %s IS NULL' % (non_visible_col, tbl, visible_col),
            'select count(%s), count(%s) from %s' % (non_visible_col, visible_col, tbl),
            # CTEs
            'SELECT * FROM ( WITH cte_non_visible_tbl as ' \
            '(SELECT %s,%s from %s) SELECT %s, %s FROM cte_non_visible_tbl) v' %
            (non_visible_col, visible_col, tbl, non_visible_col, visible_col),
            'SELECT * FROM ( WITH cte_non_visible_tbl as ' \
            '(SELECT %s,%s from %s) SELECT count(%s) FROM cte_non_visible_tbl ' \
            'group by %s) v' %
            (non_visible_col, visible_col, tbl, non_visible_col, visible_col),

            'select %s from %s group by %s' % (non_visible_col, tbl, visible_col),
            'select %s from %s group by %s' % (visible_col, tbl, non_visible_col),

            'select %s from %s a join %s b on a.%s = b.%s' %\
                 (visible_col, tbl, tbl, non_visible_col, non_visible_col),

            'select %s from %s order by %s' % (non_visible_col, tbl, visible_col),
            'select %s from %s order by %s' % (visible_col, tbl, non_visible_col),
        ]

        # Try a field in the struct if this is a struct type. These should all fail
        if non_visible_col_primitive and \
              self.__is_complex_type(types[non_visible_col][0]):
            queries += [
                'select %s from %s' % (non_visible_col_primitive, tbl),
                'select %s from %s WHERE %s IS NULL' % \
                    (visible_col, tbl, non_visible_col_primitive),
                'select %s from %s WHERE %s IS NULL' % \
                    (non_visible_col_primitive, tbl, visible_col),
                'select count(%s), count(%s) from %s' % \
                    (non_visible_col_primitive, visible_col, tbl),

                'select %s from %s group by %s' % \
                    (non_visible_col_primitive, tbl, visible_col),
                'select %s from %s group by %s' % \
                    (visible_col, tbl, non_visible_col_primitive),

                'select %s from %s order by %s' % \
                    (non_visible_col_primitive, tbl, visible_col),

                'select %s from %s order by %s' % \
                    (visible_col, tbl, non_visible_col_primitive),

                'select %s from %s a join %s b on a.%s = b.%s' %\
                    (visible_col, tbl, tbl, non_visible_col_primitive,
                     non_visible_col_primitive),
            ]

        if visible_cols:
            queries += [
                'select %s from %s a join %s b on a.%s = b.%s' %\
                    (visible_col, tbl, tbl, non_visible_col, visible_col),
                'select %s from %s a join %s b on a.%s = b.%s' %\
                    (visible_col, tbl, tbl, visible_col, non_visible_col),
            ]
            if visible_col_primitive:
                queries += [
                    'select a.%s from %s a join %s b on a.%s = b.%s' %\
                        (non_visible_col, tbl, tbl, visible_col_primitive,
                         visible_col_primitive),
                ]

        # Now run all the queries
        for query in queries:
            self.__test_invalid_query(conn, query)

        # Add this inaccessible column to the accessible ones, should fail.
        # Try a few permutations.
        select_list = visible_cols.copy()
        select_list.append(non_visible_col)
        for _ in range(0, min(len(select_list), 3)):
            random.shuffle(select_list)
            self.__test_invalid_query(
                conn, 'select %s from %s' % (','.join(select_list), tbl))

        primitive = self.__first_primitive_type(non_visible_col, types[non_visible_col])
        is_complex = self.__is_complex_type(types[non_visible_col][0])
        if primitive and is_complex and \
              user not in [ABAC_TRANSFORM_USER, ABAC_TRANSFORM_USER2]:
            # Test with a where clause on inaccessible column.
            self.__test_invalid_query(
                conn,
                'select %s from %s WHERE %s IS NOT NULL' % (visible, tbl, primitive))

    # Authorization has been configured on db.ds for visible and non-visible columns
    # on this dataset, for this user.
    # This function permutes over queries on this dataset.
    def __test_dataset_auth(self, ctx, conn, db, ds, all_cols, types,
                            visible_cols, non_visible_cols, transformed_cols,
                            user, minimal, test_scan):
        print("\nTest on %s.%s with visible cols (%s), non-visible cols(%s) for: %s" %\
              (db, ds, ', '.join(visible_cols), ', '.join(non_visible_cols), user))

        # Admin user should always see all columns
        ctx.disable_auth()
        schema = conn.plan('select * from %s.%s' % (db, ds)).schema
        result_cols, _ = self.__top_level_columns(schema)
        self.assertEqual(len(all_cols), len(result_cols))

        # Switch to test user
        ctx.enable_token_auth(token_str=user)
        if visible_cols:
            self.__test_visible_cols(conn, db, ds, user, types, visible_cols,
                                     transformed_cols, test_scan)
        else:
            self.__test_no_access(conn, db, ds)

        if minimal:
            if non_visible_cols:
                non_visible_cols_copy = non_visible_cols.copy()
                random.shuffle(non_visible_cols_copy)
                n = len(non_visible_cols_copy)
                if minimal:
                    n = min(n, MINIMAL_NUM_COLS)
                for idx in range(0, n):
                    self.__test_partial_access(
                        conn, db, ds, user, types, visible_cols,
                        non_visible_cols_copy[idx])
        else:
            for non_visible_col in non_visible_cols:
                self.__test_partial_access(
                    conn, db, ds, user, types, visible_cols, non_visible_col)

    # Runs ACL tests against db.ds
    # This test will permute through multiple dimensions to generate test cases.
    #
    # 1) ABAC/RBAC for column/table visibility
    #    X
    # 2) Queries to run
    #    X
    # 3) Clients to run it against
    #
    # This function is handles 1).
    # If minimal is true, don't exhaustive test all columns to reduce test times.
    def __test_dataset(self, ctx, conn, db, ds, minimal=False, test_scan=True):
        print("\n\nTesting dataset: %s.%s..." % (db, ds))

        ctx.disable_auth()
        print('Running valid query: select * from %s.%s' % (db, ds))
        schema = conn.plan('select * from %s.%s' % (db, ds)).schema
        all_cols, types = self.__top_level_columns(schema)
        self.__cleanup(conn)

        # Generate the columns the RBAC and ABAC user is allowed to access. The RBAC
        # user starts with no access and gets it column by column. The ABAC user does
        # the opposite.
        # TODO: add another user that uses both
        rbac_visible_cols = []
        rbac_non_visible_cols = all_cols.copy()
        abac_non_visible_cols = []
        abac_visible_cols = all_cols.copy()
        transformed_cols = []

        # Test ABAC: GRANT SELECT on table based on attribute (not in)
        self.__ddl(
            conn,
            ('GRANT SELECT ON TABLE %s.%s HAVING ATTRIBUTE ' +\
            'not in (abac_auth_test.v1) TO ROLE %s') % (db, ds, ABAC_TEST_ROLE))

        # Test transform: GRANT but only with transform. Every column is set to zero.
        self.__ddl(
            conn,
            ('GRANT SELECT ON TABLE %s.%s ' +\
            'HAVING ATTRIBUTE abac_auth_test.v1 ' + \
            'TRANSFORM abac_auth_test.v1 with zero() ' + \
            'TO ROLE %s') % (db, ds, ABAC_TRANSFORM_ROLE))

        # Test another transform GRANT. In this case the user can always see all
        # columns, and an increasing number of them are masked
        self.__ddl(
            conn,
            ('GRANT SELECT ON TABLE %s.%s ' +\
            'TRANSFORM abac_auth_test.v1 with zero() ' + \
            'TO ROLE %s') % (db, ds, ABAC_TRANSFORM_ROLE2))

        # Only admin and ABAC user can see all
        self.__test_dataset_auth(ctx, conn, db, ds, all_cols, types, rbac_visible_cols,
                                 rbac_non_visible_cols, [], TEST_USER, minimal, test_scan)
        self.__test_dataset_auth(ctx, conn, db, ds, all_cols, types, abac_visible_cols,
                                 abac_non_visible_cols, [], ABAC_TEST_USER, minimal,
                                 test_scan)
        self.__test_dataset_auth(ctx, conn, db, ds, all_cols, types, rbac_visible_cols,
                                 rbac_non_visible_cols, transformed_cols,
                                 ABAC_TRANSFORM_ROLE, minimal, test_scan)
        self.__test_dataset_auth(ctx, conn, db, ds, all_cols, types, all_cols, [],
                                 transformed_cols, ABAC_TRANSFORM_USER2, minimal,
                                 test_scan)

        # Loop over columns in this dataset. Randomize order of columns. We tend to put
        # the types in the same order across tables
        random.shuffle(all_cols)
        n = len(all_cols)
        if minimal:
            n = min(n, MINIMAL_NUM_COLS)

        for idx in range(0, n):
            c = all_cols[idx]

            # RBAC grant granting to additional columns
            ctx.disable_auth()
            self.__ddl(conn, 'GRANT SELECT(%s) ON TABLE %s.%s TO ROLE %s' %\
                  (c, db, ds, TEST_ROLE))
            rbac_visible_cols.append(c)
            rbac_non_visible_cols.remove(c)

            # Keep assigning attributes one by one that would result in exclusion of
            # the column due to ABAC evaluation. Don't cascade so each test case can
            # do its own thing.
            print("\nAssigning tag to %s.%s.%s" % (db, ds, c))
            conn.assign_attribute('abac_auth_test', 'v1', db, ds, c.strip('`'),
                                  cascade=False)
            abac_non_visible_cols.append(c)
            transformed_cols.append(c.strip('`'))
            abac_visible_cols.remove(c)

            # Grants are set up, now verify this dataset
            self.__test_dataset_auth(ctx, conn, db, ds, all_cols, types,
                                     rbac_visible_cols, rbac_non_visible_cols, [],
                                     TEST_USER, minimal, test_scan)
            self.__test_dataset_auth(ctx, conn, db, ds, all_cols, types,
                                     abac_visible_cols, abac_non_visible_cols, [],
                                     ABAC_TEST_USER, minimal, test_scan)
            self.__test_dataset_auth(ctx, conn, db, ds, all_cols, types,
                                     rbac_visible_cols, rbac_non_visible_cols,
                                     transformed_cols,
                                     ABAC_TRANSFORM_USER, minimal, test_scan)
            self.__test_dataset_auth(ctx, conn, db, ds, all_cols, types, all_cols, [],
                                     transformed_cols, ABAC_TRANSFORM_USER2, minimal,
                                     test_scan)

        # Grant to entire table and verify
        ctx.disable_auth()
        self.__ddl(conn, 'GRANT SELECT ON TABLE %s.%s TO ROLE %s' %\
              (db, ds, TEST_ROLE))
        self.__test_dataset_auth(ctx, conn, db, ds, all_cols, types,
                                 all_cols, [], [], TEST_USER, minimal, test_scan)

    # Tests that verify grants to complex types works. This programmatically gets
    # columns to the table/view and increasingly grants more access to it.
    @unittest.skipIf(common.TEST_LEVEL in ["all"], "Skipping at all level")
    def test_complex_types_core(self):
        # This contains a subset that capture most cases
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            for ds in ['struct_t', 'strarray_t_view']:
                self.__test_dataset(ctx, conn, 'rs_complex', ds)

    @unittest.skipIf(common.TEST_LEVEL in SKIP_LEVELS, "Skipping at unit/all level")
    def test_complex_types_extended(self):
        # This adds some additional tables and run at a higher test level to control
        # test times.
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            for ds in ['struct_t2', 'struct_t3', 'struct_array_struct', 'struct_view',
                       'struct_t_restricted_view', 'struct_nested', 'struct_t_view2',
                       'struct_nested_view', 'array_struct_array', 'array_struct_t',
                       'array_t', 'avrotbl', 'map_t', 'multiple_structs_nested', 'users',
                       'user_phone_numbers', 'user_phone_numbers_map']:
                self.__test_dataset(ctx, conn, 'rs_complex', ds)

    @unittest.skipIf(common.TEST_LEVEL in SKIP_LEVELS, "Skipping at unit/all level")
    def test_market_data(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self.__test_dataset(ctx, conn, 'rs_complex', 'market_v20_single')

    @unittest.skipIf(common.TEST_LEVEL in SKIP_LEVELS, "Skipping at unit/all level")
    def test_complex_types_slow(self):
        # These tests are N^2 with the number of columns so don't run these by default
        slow = ['test_bucketing_tbl', 'zd1216', 'zd1216_with_subscriptionlimit']

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            for ds in slow:
                self.__test_dataset(ctx, conn, 'rs_complex', ds, True)
            for ds in ['ledger_balance', 'subscription', 'subscription_currency',
                       'subscription_view']:
                self.__test_dataset(ctx, conn, 'chase', ds, True)

    @staticmethod
    def check_dataset_metadata(database, dataset, prop, value):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            ds_info = conn.list_datasets(db=database, name=dataset)
            assert(len(ds_info) == 1)
            ds_md = ds_info[0].metadata
            return (ds_md is not None) and (prop in ds_md) and (ds_md[prop] == value)

    def __test_db(self, ctx, conn, db, minimal):
        print("Testing database: " + db)
        ctx.disable_auth()
        datasets = conn.list_dataset_names(db)
        for ds in datasets:
            # If set to false, don't run the scanning (just the planning). We do
            # for big datasets that can take a long time to to skip tables that
            # we can't scan due to bugs (but still get some test coverage).
            run_scan = True

            # These datasets/dbs take a long time to run as the dataset is big
            if ds in ['buckets_test.nytaxi_sample_bucketed',
                      'nytaxi.orc_data',
                      'nytaxi.parquet_data',
                      'nytaxi.parquet_mega',
                      'nytaxi.parquet_data_symlink',
                      'nytaxi.unionallview',
                      'partition_test.part_flat_200_data_100',
                      'rs.alltypes_large_s3',
                      'rs.alltypes_large_s3_gz',
                      'rs.alltypes_large_s3_gz_no_ext',
                      'rs.alltypes_large_s3_lzo']:
                run_scan = False

            if db in ['tpcds1000_partitioned', 'tpcds1000_unpartitioned',
                      'tpcds100_partitioned', 'tpcds100_unpartitioned',
                      'tpcds10_partitioned', 'tpcds10_unpartitioned',
                      'tpcds1_partitioned', 'tpcds1_unpartitioned',
                      'tpcds_csv_10', 'tpcds_csv_100', 'tpch_sf1', 'tpch_sf5',
                      'zd1179large']:
                run_scan = False

            # Complex types on text format, not supported.
            # This fails to plan.
            if ds in ['customer.japan_pos_trade_item_stg',
                      'customer.japan_pos_trade_item_stg_part',
                      'rs_complex.array_text',
                      'rs_complex.map_text']:
                continue

            # Data is inaccessible in s3. This table is not suppose to work.
            # This fails to plan.
            if ds in ['rs.s3_no_perm']:
                continue

            # Data stored in DBFS which we can't plan.
            # This fails to plan.
            if ds in ['customer.zd1010_dbfs_table_external_provider',
                      'customer.zd1065_dbfs_table_no_provider']:
                continue

            # BUG[DAS-4871]
            # Test fail to assign tags due to special characters in the column name.
            # This fails to plan.
            if ds in ['special_chars.space_table',
                      'special_chars.test_table']:
                continue

            # BUG: Self join bug
            # This fails to plan.
            if ds in ['customer.zd608_fiducia_risk_score_ato_tb',
                      'demo.reporting_audit_logs',
                      'partition_test.special_chars_partition',
                      'partition_test.special_chars_partition_nested',
                      'partition_test.timestamp_part_encoded',
                      'partition_test.timestamp_part_test',
                      'partition_test.timestamp_part_test2']:
                continue

            # BUG[DAS-4799]: This crashes the worker
            # This fails to scan.
            if ds in ['parquet_testing.dict_page_offset_zero']:
                run_scan = False

            # BUG[DAS-5908]: This crashes the worker
            # This fails to scan.
            if ds in ['rs_complex.unnest1', 'rs_complex.unnest2']:
                run_scan = False

            # BUG[DAS-4801]: Worker fails this query with nested array items
            # This fails to scan.
            if ds in ['parquet_testing.nested_lists_snappy',
                      'parquet_testing.nonnullable_impala',
                      'parquet_testing.nullable_impala',
                      'rs_complex_parquet.rs_parquet_array_map_t',
                      'rs_complex.rs_complex_array_map_t',]:
                run_scan = False

            # BUG[DAS-4802]: pyokera fails to deserialize with index out of bounds
            # This fails to scan.
            if ds in ['parquet_testing.nested_maps_snappy']:
                run_scan = False

            # BUG[DAS-4905]: This where clause is not supported right and dropped
            # This fails to scan.
            if ds in ['chase.subscription_party_view']:
                continue

            # BUG[DAS-5500]: Fail to convert
            if ds in ['customer.t2_authzn', 'partition_test.keyword_part_table',
                      'datedb.dates_with_invalid_data', 'partition_test.weird_partition6',
                      'rs.partitioned_tbl_s3']:
                continue

            # Bad avro format cant be scanned
            if ds in ['customer.zd623_east', 'rs.avro_comments']:
                run_scan = False

            # Unsupported
            if ds in ['parquet_testing.datapage_v2_snappy', 'rs_complex.pn_view',
                      'parquet_testing.nested_maps_snappy',
                      'parquet_testing.repeated_no_annotation']:
                continue


            # JDBC datasets
            if ds in ['jdbc_psql_test.nytaxi_demo',
                      'jdbc_test_athena._abac_privileges',
                      'jdbc_test_athena._sentry_privileges',
                      'jdbc_test_athena.privileges']:
                continue

            # Skip if ds is external view
            if self.check_dataset_metadata(ds.split('.')[0], ds.split('.')[1],
                                           'cerebro.external.view', 'true'):
                continue

            self.__test_dataset(ctx, conn, db, ds.split('.')[1], minimal, run_scan)

    @unittest.skip("Example for use during development to test a specific dataset.")
    def test_single_dataset(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self.__test_dataset(
                ctx, conn, 'rs_complex', 'struct_view', minimal=False, test_scan=True)

    @unittest.skipIf(common.TEST_LEVEL in SKIP_LEVELS, "Skipping at unit/all level")
    def test_rs_db(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            for ds in ['alltypes', 'alltypes_null', 'ccn',
                       'large_decimals', 'nation_projection',
                       's3_nation', 'users']:
                self.__test_dataset(ctx, conn, 'rs', ds)

    @unittest.skipIf(common.TEST_LEVEL in SKIP_LEVELS, "Skipping at unit/all level")
    def test_auth_db(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self.__test_db(ctx, conn, 'authdb', True)

    @unittest.skipIf(common.TEST_LEVEL != "all", "Skipping at unit test level")
    def test_all(self):
        # Runs test again all tables in all dbs, except the black listed ones
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            dbs = conn.list_databases()
            for db in dbs:
                # Dbs that either have public access or non queryable tables
                # BUG DAS-5544 filter pushdown on jdbc tables
                # Delta tables are getting errors from the manifest file when scanning
                if db in ['bad_metadata_db', 'cerebro_sample', 'default', 'demo',
                          'okera_sample', 'okera_system', 'all_table_types',
                          'jdbc_test_mysql', 'jdbc_test_psql', 'jdbc_test_redshift',
                          'jdbc_test_snowflake', 'jdbc_test_sqlserver',
                          'jdbc_test_oracle', 'delta_db']:
                    continue
                if db.startswith('_'):
                    continue

                self.__test_db(ctx, conn, db, True)

if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Okera Inc. All Rights Reserved.
#
# Tests for ORC support
# pylint: disable=bad-continuation,bad-indentation,global-statement,unused-argument
# pylint: disable=no-self-use
import unittest

from okera.tests import pycerebro_test_common as common
import cerebro_common as cerebro

DB = "orc_test_db"

class OrcTest(common.TestBase):
    def test_binary_data(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            tbl = "%s.%s" % (DB, "binary_data")
            ddls = [
                "DROP DATABASE IF EXISTS %s CASCADE" % (DB),
                "CREATE DATABASE %s" % (DB),
                """CREATE EXTERNAL TABLE %s (
                    bytes1 binary,
                    string1 string
                )
                STORED AS ORC
                LOCATION 's3://cerebrodata-test/orc/binary_data/'""" % (tbl),
            ]

            for ddl in ddls:
                conn.execute_ddl(ddl)

            results = conn.scan_as_json(tbl)

            # Extracted from https://github.com/apache/orc/blob/7de5d89/examples/expected/TestOrcFile.testStringAndBinaryStatistics.jsn.gz
            expected_results = [
                {'string1': 'foo', 'bytes1': '\x00\x01\x02\x03\x04'},
                {'string1': 'bar', 'bytes1': '\x00\x01\x02\x03'},
                {'string1': None, 'bytes1': '\x00\x01\x02\x03\x04\x05'},
                {'string1': 'hi', 'bytes1': None}]

            assert results == expected_results

if __name__ == "__main__":
    unittest.main()



import requests
import os
import pprint
import json

def get_okera_token():
    headers = {
        'User-Agent': 'okera-token-generator-loopback/jupyter',
        'Authorization': 'token ' + os.environ.get('JUPYTERHUB_API_TOKEN')
    }
    r = requests.get(os.environ.get('OKERA_TOKEN_GENERATOR_ENDPOINT'), headers=headers, verify=False)
    result = json.loads(r.content)
    return result.get('okera_token')


# Copyright 2017 Okera Inc. All Rights Reserved.
#
# Tests that verify connection issues with specific server auth settings.

# pylint: disable=broad-except
# pylint: disable=bad-continuation
# pylint: disable=bad-indentation
import os
import unittest

from okera.tests import pycerebro_test_common as common

SERVER_AUTH = 'NOSASL'
if 'PYCEREBRO_TEST_AUTH_MECH' in os.environ:
    SERVER_AUTH = os.environ['PYCEREBRO_TEST_AUTH_MECH']

ROOT_TOKEN = os.environ['OKERA_HOME'] + '/integration/tokens/cerebro.token'

class ConnectionErrorsTest(unittest.TestCase):

    @unittest.skipIf(SERVER_AUTH != 'NOSASL', "Test requires server to have no auth.")
    def test_unauthenticated_server_planner(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth('bad.token')
        thrown = False
        try:
            common.get_planner(ctx)
        except Exception as e:
            msg = str(e)
            self.assertTrue("Ensure server has authentication enabled" in msg, msg=e)
            thrown = True
        self.assertTrue(thrown)

    @unittest.skipIf(SERVER_AUTH != 'NOSASL', "Test requires server to have no auth.")
    def test_unauthenticated_server_worker(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth('bad.token')
        thrown = False
        try:
            common.get_worker(ctx)
        except Exception as e:
            msg = str(e)
            self.assertTrue("Ensure server has authentication enabled" in msg, msg=e)
            thrown = True
        self.assertTrue(thrown)

    @unittest.skipIf(SERVER_AUTH != 'TOKEN', "Test requires server to have token auth.")
    def test_token_server_wrong_token(self):
        ctx = common.get_test_context()

        # Wrong token
        thrown = False
        try:
            ctx.enable_token_auth(token_str='bad.token')
            common.get_planner(ctx)
        except Exception as e:
            msg = str(e)
            thrown = True
            self.assertTrue("Could not verify token" in msg, msg=e)
        self.assertTrue(thrown)

    @unittest.skipIf(SERVER_AUTH != 'TOKEN', "Test requires server to have token auth.")
    def test_token_server_no_token(self):
        thrown = False
        ctx = common.get_test_context()
        try:
            ctx.disable_auth()
            common.get_planner(ctx)
        except Exception as e:
            msg = str(e)
            thrown = True
            self.assertTrue("Client does not have authentication enabled" in msg, msg=e)
        self.assertTrue(thrown)

    @unittest.skipIf(SERVER_AUTH != 'TOKEN', "Test requires server to have kerberos.")
    def test_token_server_valid_token(self):
        ctx = common.get_test_context()
        ctx.enable_token_auth(token_file=ROOT_TOKEN)
        with common.get_planner(ctx) as p:
            self.assertTrue(p.get_protocol_version() is not None)

    @unittest.skipIf(SERVER_AUTH != 'KERBEROS', "Test requires server to have kerberos.")
    def test_kerberos_server_no_override(self):
        ctx = common.get_test_context()
        # This assumes the server principal is cerebro/localhost, which is wrong
        ctx.enable_kerberos('cerebro')
        thrown = False
        try:
            common.get_planner(ctx)
        except Exception as e:
            msg = str(e)
            thrown = True
            self.assertTrue("cerebro/localhost@CEREBRO.TEST" in msg, msg=e)
        self.assertTrue(thrown)

    @unittest.skipIf(SERVER_AUTH != 'KERBEROS', "Test requires server to have kerberos.")
    def test_kerberos_server_valid_client(self):
        ctx = common.get_test_context()
        ctx.enable_kerberos('cerebro', host_override='service')
        with common.get_planner(ctx) as p:
            self.assertTrue(p.get_protocol_version() is not None)
            p.close()

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# Copyright Okera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import ez_setup
ez_setup.use_setuptools()

import sys
from os.path import dirname, exists, join
import os
from setuptools import setup, find_packages, Extension
import versioneer  # noqa

if not exists((join(dirname(__file__), 'okera', 'thrift', 'RecordService.thrift'))):
    print("thrift files not present. Run build.sh from this folder.")
    sys.exit(1)

def readme():
    with open('README.rst', 'r') as ip:
        readme = ip.read()
        if 'PUBLISH_VERSION' in os.environ:
            return readme.replace('##OKERA_RELEASE_VERSION##', os.environ['PUBLISH_VERSION'])
        return readme

reqs = [
    'six', 'bitarray!=1.6.2', 'certifi', 'thriftpy2==0.3.12',
    'urllib3', 'pytz', 'jwt', 'presto-python-client']

setup(
    name='pyokera',
    maintainer='Okera Development Team',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Python client for the Okera Data Access Service',
    long_description=readme(),
    packages = find_packages(exclude=['okera.RecordService',
                                      'okera.RecordService.*',
                                      'okera.OkeraRecordService',
                                      'okera.OkeraRecordService.*']),
    package_data={'okera': ['thrift/*.thrift']},
    install_requires=reqs,
    license='Apache License, Version 2.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    zip_safe=False)



# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests for role granting
#
# pylint: disable=too-many-public-methods,no-self-use

import unittest

from okera import _thrift_api
from okera.tests import pycerebro_test_common as common
from okera._thrift_api import (
    TAccessPermissionLevel,
    TAccessPermissionScope
)

TEST_USER = 'roles_test_user'
TEST_USER2 = 'roles_test_user2'

OKERA_PUBLIC_ROLE = 'okera_public_role'

def disable_auth(ctx, params):
    ctx.disable_auth()
    params.requesting_user = None

def enable_auth(ctx, params, user):
    ctx.enable_token_auth(token_str=user)
    params.requesting_user = user

def get_grantable_roles(conn, params):
    #pylint: disable=protected-access
    return conn._underlying_client().GetGrantableRoles(params)
    #pylint: enable=protected-access

def get_role(conn, name):
    #pylint: disable=protected-access
    params = _thrift_api.TGetRoleParams()
    params.role_name = name
    if conn.ctx._get_user():
        params.requesting_user = conn.ctx._get_user()
    return conn._underlying_client().GetRole(params)
    #pylint: enable=protected-access

def list_roles(conn, params, limit=None):
    #pylint: disable=protected-access
    if conn.ctx._get_user():
        params.requesting_user = conn.ctx._get_user()
    if limit is not None:
        params.limit = limit
    else:
        params.limit = 0
    return conn._underlying_client().ListRoles(params)
    #pylint: enable=protected-access

def contains_role(role_name, roles):
    for role in roles:
        if role.role_name == role_name:
            return True

    return False

def get_role_by_name(role_name, roles):
    for role in roles:
        if role.role_name == role_name:
            return role

    return None

# def cleanup_for_user(conn, user):
#     conn.ctx.enable_token_auth(conn)
#     roles_for_user

class RolesTest(unittest.TestCase):
    # NOTE: this test will likely need to change once we can properly
    # ACL roles themselves, since then we will return just a subset of
    # roles.
    def test_get_grantable_roles(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            params = _thrift_api.TGetGrantableRolesParams()

            # Set up a role and grant it to the user's group
            disable_auth(ctx, params)
            conn.execute_ddl("DROP ROLE IF EXISTS grantable_role")
            conn.execute_ddl("CREATE ROLE grantable_role")
            conn.execute_ddl("GRANT ROLE grantable_role to GROUP %s" % TEST_USER)

            # As an admin, we should always be able to see all roles
            disable_auth(ctx, params)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) > 0)
            self.assertTrue("grantable_role" in retrieved_roles)

            # Even though we have the test role, our user does not have
            # any roles that have a privilege with GRANT OPTION, so we
            # should not get back any roles
            enable_auth(ctx, params, TEST_USER)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) == 0)

            # As an admin, grant a privilege with GRANT OPTION to the
            # role we created
            disable_auth(ctx, params)
            conn.execute_ddl("""
            GRANT SELECT ON TABLE okera_sample.sample
            TO ROLE grantable_role WITH GRANT OPTION""")
            admin_roles = get_grantable_roles(conn, params).roles

            # Even though there is a grant with GRANT OPTION, it's on
            # a specific table, and we specify no parameters, which is
            # equivalent to searching for grants on the CATALOG
            enable_auth(ctx, params, TEST_USER)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) == 0)

            # Now, we will check with specific parameters
            params.database = "okera_sample"
            params.table = "sample"

            # As a user, we should now get back all the roles since we have a
            # privilege that has a GRANT OPTION that matches our scope. The set of
            # roles should be the same as what the admin got.
            enable_auth(ctx, params, TEST_USER)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) > 0)
            self.assertEqual(admin_roles, retrieved_roles)

            # Now, we will check with specific parameters on a table we don't have
            params.database = "okera_sample"
            params.table = "users"

            # As a user, we should get no roles, as we don't have a GRANT OPTION
            # on that specific scope
            enable_auth(ctx, params, TEST_USER)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) == 0)

            # Now, we will check with specific parameters on just the DB
            params.database = "okera_sample"

            # As a user, we should get no roles, as we don't have a GRANT OPTION
            # on that specific scope
            enable_auth(ctx, params, TEST_USER)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) == 0)

            # As an admin, grant a privilege with GRANT OPTION to the
            # role we created
            disable_auth(ctx, params)
            conn.execute_ddl("""
            GRANT SELECT ON DATABASE okera_sample
            TO ROLE grantable_role WITH GRANT OPTION""")

            # As a user, we should now get all roles, as we have a GRANT OPTION
            # on that specific scope
            enable_auth(ctx, params, TEST_USER)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) > 0)
            self.assertEqual(admin_roles, retrieved_roles)

            # Now, we will check with specific parameters on just the DB,
            # with a different DB
            params.database = "okera_system"

            # As a user, we should get no roles, as we don't have a GRANT OPTION
            # on that specific scope
            enable_auth(ctx, params, TEST_USER)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) == 0)

            # Finally, let's add a grant to the CATALOG and then
            # a query with no parameters will work
            disable_auth(ctx, params)
            conn.execute_ddl("""
            GRANT SELECT ON SERVER
            TO ROLE grantable_role WITH GRANT OPTION""")

            # Now, we will check with no parameters
            params.database = ""
            params.table = ""

            # As a user, we should get all the roles, we have GRANT OPTION
            # at the CATALOG level
            enable_auth(ctx, params, TEST_USER)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) > 0)
            self.assertEqual(admin_roles, retrieved_roles)

            # Additionally, even if we query on a specific DB (or even table),
            # we will now always get roles because our CATALOG level grant
            # allows us to see them.
            params.database = "okera_system"

            enable_auth(ctx, params, TEST_USER)
            response = get_grantable_roles(conn, params)
            retrieved_roles = response.roles
            self.assertTrue(len(retrieved_roles) > 0)
            self.assertEqual(admin_roles, retrieved_roles)

    def test_drop_privilege(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            obj_name = "drop_privilege_obj"
            role_name = "drop_privilege_role"

            # TODO: when we add ACLs on roles, should add a test here
            ddls = [
                # Setup role
                "DROP ROLE IF EXISTS %s" % (role_name),
                "CREATE ROLE %s" % (role_name),
                "GRANT ROLE %s TO GROUP %s" % (role_name, TEST_USER),

                # DB
                "DROP DATABASE IF EXISTS %s CASCADE" % (obj_name),
                "CREATE DATABASE %s" % (obj_name),
                "GRANT DROP ON DATABASE %s TO ROLE %s" % (obj_name, role_name),

                # Table
                "CREATE TABLE %s.%s (c1 string)" % (obj_name, obj_name),
                "GRANT DROP ON TABLE %s.%s TO ROLE %s" % (obj_name, obj_name, role_name),

                # View
                "CREATE VIEW %s.%s AS SELECT 1" % (obj_name, obj_name+"view"),
                "GRANT DROP ON TABLE %s.%s TO ROLE %s" % (obj_name,
                                                          obj_name+"view", role_name),

                # Attribute Namespace
                "DROP ATTRIBUTE IF EXISTS %s.%s" % (obj_name, obj_name),
                "CREATE ATTRIBUTE %s.%s" % (obj_name, obj_name),
                "GRANT DROP ON ATTRIBUTE NAMESPACE %s TO ROLE %s" % (obj_name, role_name),
            ]

            for ddl in ddls:
                conn.execute_ddl(ddl)

            # Switch to the user
            ctx.enable_token_auth(token_str=TEST_USER)

            # Verify we can drop everything (we do it in reverse order from the DDLs
            # to ensure we can drop the DB at the end, as it needs to be empty)
            drop_ddls = [
                "DROP ATTRIBUTE %s.%s" % (obj_name, obj_name),
                "DROP VIEW %s.%s" % (obj_name, obj_name+"view"),
                "DROP TABLE %s.%s" % (obj_name, obj_name),
                "DROP DATABASE %s" % (obj_name),
            ]

            for drop_ddl in drop_ddls:
                conn.execute_ddl(drop_ddl)

    def test_role_priv_drop(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            steward_role = "role_priv_steward_role"
            obj_role = "role_priv_test_role"

            for obj in ['CATALOG', 'ROLE %s' % obj_role]:
                for (priv, succeed) in [('DROP', True), ('ALL', True), ('MANAGE_GROUPS', False)]:
                    ctx.disable_auth()
                    setup_ddls = [
                        # Setup steward role
                        "DROP ROLE IF EXISTS %s" % (steward_role),
                        "CREATE ROLE %s" % (steward_role),
                        "GRANT ROLE %s TO GROUP %s" % (steward_role, TEST_USER),

                        # Setup object role
                        "DROP ROLE IF EXISTS %s" % (obj_role),
                        "CREATE ROLE %s" % (obj_role),

                        # Grant specific privilege on specified object to the
                        # steward role
                        "GRANT %s ON %s TO ROLE %s" % (priv, obj, steward_role),
                    ]

                    for ddl in setup_ddls:
                        conn.execute_ddl(ddl)

                    # Switch to the user
                    ctx.enable_token_auth(token_str=TEST_USER)

                    # Verify we can drop the role
                    print("Trying to DROP ROLE with GRANT %s ON %s - should %s" % (
                        priv, obj, 'succeed' if succeed else 'fail'))

                    if succeed:
                        conn.execute_ddl('DROP ROLE %s' % obj_role)
                    else:
                        with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                            conn.execute_ddl('DROP ROLE %s' % obj_role)
                        self.assertTrue('to DROP this role' in str(ex_ctx.exception))

    def test_role_priv_create(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            steward_role = "role_priv_steward_role"
            obj_role = "role_priv_test_role"
            as_owner_role = "_okera_internal_role_%s" % (TEST_USER)

            # (privilege, should succeed, as_owner semantics)
            combinations = [
                ('ALL', True, False),
                ('CREATE_ROLE_AS_OWNER', True, True),
                ('CREATE', False, False),
            ]

            for obj in ['CATALOG']:
                for (priv, succeed, as_owner) in combinations:
                    ctx.disable_auth()
                    setup_ddls = [
                        # Setup steward role
                        "DROP ROLE IF EXISTS %s" % (steward_role),
                        "CREATE ROLE %s" % (steward_role),
                        "GRANT ROLE %s TO GROUP %s" % (steward_role, TEST_USER),

                        # Cleanup the object role
                        "DROP ROLE IF EXISTS %s" % (obj_role),

                        # Cleanup the special per-user role to start from a clean slate
                        "DROP ROLE IF EXISTS %s" % (as_owner_role),

                        # Grant specific privilege on specified object to the
                        # steward role
                        "GRANT %s ON %s TO ROLE %s" % (priv, obj, steward_role),
                    ]

                    for ddl in setup_ddls:
                        conn.execute_ddl(ddl)

                    # Switch to the user
                    ctx.enable_token_auth(token_str=TEST_USER)

                    # Verify we can drop the role
                    print("Trying to CREATE ROLE with GRANT %s ON %s - should %s" % (
                        priv, obj, 'succeed' if succeed else 'fail'))

                    if succeed:
                        conn.execute_ddl('CREATE ROLE %s' % obj_role)
                    else:
                        with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                            conn.execute_ddl('CREATE ROLE %s' % obj_role)
                        self.assertTrue('to CREATE this role' in str(ex_ctx.exception))

                    if as_owner:
                        role_grants = conn.execute_ddl('SHOW GRANT ROLE %s' % as_owner_role)
                        print(role_grants)
                        assert len(role_grants) == 1
                        assert role_grants[0][0] == 'role'
                        assert role_grants[0][6] == obj_role
                        assert role_grants[0][9] == 'all'
                        assert role_grants[0][13] == 'false'
                    else:
                        with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                            conn.execute_ddl('SHOW GRANT ROLE %s' % as_owner_role)
                        self.assertTrue('does not exist' in str(ex_ctx.exception))

    def test_role_priv_manage_groups(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            steward_role = "role_priv_steward_role"
            obj_role = "role_priv_test_role"

            for obj in ['CATALOG', 'ROLE %s' % obj_role]:
                for (priv, succeed) in [('MANAGE_GROUPS', True), ('ALL', True), ('DROP', False)]:
                    ctx.disable_auth()
                    setup_ddls = [
                        # Setup steward role
                        "DROP ROLE IF EXISTS %s" % (steward_role),
                        "CREATE ROLE %s" % (steward_role),
                        "GRANT ROLE %s TO GROUP %s" % (steward_role, TEST_USER),

                        # Setup object role
                        "DROP ROLE IF EXISTS %s" % (obj_role),
                        "CREATE ROLE %s" % (obj_role),

                        # Grant specific privilege on specified object to the
                        # steward role
                        "GRANT %s ON %s TO ROLE %s" % (priv, obj, steward_role),
                    ]

                    for ddl in setup_ddls:
                        conn.execute_ddl(ddl)

                    # Switch to the user
                    ctx.enable_token_auth(token_str=TEST_USER)

                    # Verify we can grant the role to a group
                    print("Trying to GRANT ROLE TO GROUP with GRANT %s ON %s - should %s" % (
                        priv, obj, 'succeed' if succeed else 'fail'))

                    if succeed:
                        conn.execute_ddl('GRANT ROLE %s TO GROUP some_group' % obj_role)
                    else:
                        with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                            conn.execute_ddl('GRANT ROLE %s TO GROUP some_group' % obj_role)
                        self.assertTrue('GRANT/REVOKE this role to groups.' in str(ex_ctx.exception))

                    # Verify we can revoke the role from a group
                    print("Trying to REVOKE ROLE FROM GROUP with GRANT %s ON %s - should %s" % (
                        priv, obj, 'succeed' if succeed else 'fail'))

                    if succeed:
                        conn.execute_ddl('REVOKE ROLE %s FROM GROUP some_group' % obj_role)
                    else:
                        with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                            conn.execute_ddl('REVOKE ROLE %s FROM GROUP some_group' % obj_role)
                        self.assertTrue('GRANT/REVOKE this role to groups.' in str(ex_ctx.exception))

    def test_role_priv_manage_permissions(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            steward_role = "role_priv_steward_role"
            obj_role = "role_priv_test_role"
            test_db = "role_priv_manage_permissions_db"

            for obj in ['CATALOG', 'ROLE %s' % obj_role]:
                for (priv, succeed) in [('MANAGE_PERMISSIONS', True), ('ALL', True), ('MANAGE_GROUPS', False)]:
                    ctx.disable_auth()
                    setup_ddls = [
                        # Setup DB
                        "DROP DATABASE IF EXISTS %s CASCADE" % (test_db),
                        "CREATE DATABASE %s" % (test_db),

                        # Setup steward role
                        "DROP ROLE IF EXISTS %s" % (steward_role),
                        "CREATE ROLE %s" % (steward_role),
                        "GRANT ROLE %s TO GROUP %s" % (steward_role, TEST_USER),

                        # Setup object role
                        "DROP ROLE IF EXISTS %s" % (obj_role),
                        "CREATE ROLE %s" % (obj_role),

                        # Grant specific privilege on specified object to the
                        # steward role
                        "GRANT %s ON %s TO ROLE %s" % (priv, obj, steward_role),
                        "GRANT SELECT ON DATABASE %s TO ROLE %s WITH GRANT OPTION" % (test_db, steward_role),
                    ]

                    for ddl in setup_ddls:
                        conn.execute_ddl(ddl)

                    # Switch to the user
                    ctx.enable_token_auth(token_str=TEST_USER)

                    # Verify we can grant the role to a group
                    print("Trying to 'GRANT SELECT TO ROLE' with GRANT %s ON %s - should %s" % (
                        priv, obj, 'succeed' if succeed else 'fail'))

                    if succeed:
                        conn.execute_ddl('GRANT SELECT ON DATABASE %s TO ROLE %s' % (test_db, obj_role))
                    else:
                        with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                            conn.execute_ddl('GRANT SELECT ON DATABASE %s TO ROLE %s' % (test_db, obj_role))
                        self.assertTrue('not have privileges to GRANT' in str(ex_ctx.exception))

                    # Verify we can revoke the database from the role group
                    print("Trying to 'REVOKE SELECT FROM ROLE' with GRANT %s ON %s - should %s" % (
                        priv, obj, 'succeed' if succeed else 'fail'))

                    if succeed:
                        conn.execute_ddl('REVOKE SELECT ON DATABASE %s FROM ROLE %s' % (test_db, obj_role))
                    else:
                        with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                            conn.execute_ddl('REVOKE SELECT ON DATABASE %s FROM ROLE %s' % (test_db, obj_role))
                        self.assertTrue('not have privileges to REVOKE' in str(ex_ctx.exception))

    def test_role_priv_show_roles(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            steward_role = "role_priv_steward_role"
            obj_role1 = "role_priv_test_role1"
            obj_role2 = "role_priv_test_role2"
            obj_role3 = "role_priv_test_role3"
            obj_role4 = "role_priv_test_role4"

            for level in ['SHOW', 'MANAGE_GROUPS']:
                setup_ddls = [
                    "DROP ROLE IF EXISTS grantable_role",

                    # Setup steward role
                    "DROP ROLE IF EXISTS %s" % (steward_role),
                    "CREATE ROLE %s" % (steward_role),
                    "GRANT ROLE %s TO GROUP %s" % (steward_role, TEST_USER),

                    # Setup object roles
                    "DROP ROLE IF EXISTS %s" % (obj_role1),
                    "CREATE ROLE %s" % (obj_role1),
                    "DROP ROLE IF EXISTS %s" % (obj_role2),
                    "CREATE ROLE %s" % (obj_role2),
                    "DROP ROLE IF EXISTS %s" % (obj_role3),
                    "CREATE ROLE %s" % (obj_role3),
                    "DROP ROLE IF EXISTS %s" % (obj_role4),
                    "CREATE ROLE %s" % (obj_role4),

                    # Grant the last role to us explicitly
                    "GRANT ROLE %s TO GROUP %s" % (obj_role4, TEST_USER),

                    # Grant specific privilege on specified object to the
                    # steward role
                    "GRANT %s ON ROLE %s TO ROLE %s" % (level, obj_role1, steward_role),
                    "GRANT %s ON ROLE %s TO ROLE %s" % (level, obj_role2, steward_role),
                ]

                for ddl in setup_ddls:
                    conn.execute_ddl(ddl)

                print("Running SHOW ROLES tests for access level: %s" % (level))

                # Switch to the user who has the roles
                ctx.enable_token_auth(token_str=TEST_USER)

                # SHOW ROLES should show us all the roles we have access to.
                show_roles = [x[0] for x in conn.execute_ddl('SHOW ROLES')]
                assert steward_role in show_roles
                assert obj_role1 in show_roles
                assert obj_role2 in show_roles
                assert obj_role3 not in show_roles
                assert obj_role4 in show_roles

                # SHOW CURRENT ROLES should show us all the roles we have.
                show_roles = [x[0] for x in conn.execute_ddl('SHOW CURRENT ROLES')]
                assert steward_role in show_roles
                assert obj_role1 not in show_roles
                assert obj_role2 not in show_roles
                assert obj_role3 not in show_roles
                assert obj_role4 in show_roles

                # SHOW ROLE GRANT GROUP should show us all the group has, but since
                # it's our user group, this is equivalent to `SHOW CURRENT ROLES`
                show_roles = [x[0] for x in conn.execute_ddl('SHOW ROLE GRANT GROUP %s' % TEST_USER)]
                assert steward_role in show_roles
                assert obj_role1 not in show_roles
                assert obj_role2 not in show_roles
                assert obj_role3 not in show_roles
                assert obj_role4 in show_roles

                # SHOW ALL ROLES is only accessible to admins.
                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                    show_roles = conn.execute_ddl('SHOW ALL ROLES')
                self.assertTrue("'roles_test_user' is not an admin" in str(ex_ctx.exception))

                # SHOW ROLE GRANT GROUP on a group the user is not in is
                # only accessible to admins.
                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                    show_roles = conn.execute_ddl('SHOW ROLE GRANT GROUP %s' % TEST_USER2)
                self.assertTrue("'roles_test_user' is not an admin" in str(ex_ctx.exception))

                # Switch to the other user who has no roles
                ctx.enable_token_auth(token_str=TEST_USER2)

                # SHOW ROLES should show us all the roles we have access to.
                show_roles = [x[0] for x in conn.execute_ddl('SHOW ROLES')]
                assert steward_role not in show_roles
                assert obj_role1 not in show_roles
                assert obj_role2 not in show_roles
                assert obj_role3 not in show_roles
                assert obj_role4 not in show_roles

                # SHOW CURRENT ROLES should show us all the roles we have.
                show_roles = [x[0] for x in conn.execute_ddl('SHOW CURRENT ROLES')]
                assert steward_role not in show_roles
                assert obj_role1 not in show_roles
                assert obj_role2 not in show_roles
                assert obj_role3 not in show_roles
                assert obj_role4 not in show_roles

                # SHOW ROLE GRANT GROUP should show us all the group has, but since
                # it's our user group, this is equivalent to `SHOW CURRENT ROLES`
                show_roles = [x[0] for x in conn.execute_ddl('SHOW ROLE GRANT GROUP %s' % TEST_USER2)]
                assert steward_role not in show_roles
                assert obj_role1 not in show_roles
                assert obj_role2 not in show_roles
                assert obj_role3 not in show_roles
                assert obj_role4 not in show_roles

                # SHOW ALL ROLES is only accessible to admins.
                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                    show_roles = conn.execute_ddl('SHOW ALL ROLES')
                self.assertTrue("'roles_test_user2' is not an admin" in str(ex_ctx.exception))

                # SHOW ROLE GRANT GROUP on a group the user is not in is
                # only accessible to admins.
                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                    show_roles = conn.execute_ddl('SHOW ROLE GRANT GROUP %s' % TEST_USER)
                self.assertTrue("'roles_test_user2' is not an admin" in str(ex_ctx.exception))

                # Switch to admin user
                ctx.disable_auth()

                # SHOW ROLES should show us all the roles we have access to.
                show_roles = [x[0] for x in conn.execute_ddl('SHOW ROLES')]
                assert steward_role in show_roles
                assert obj_role1 in show_roles
                assert obj_role2 in show_roles
                assert obj_role3 in show_roles
                assert obj_role4 in show_roles

                # SHOW CURRENT ROLES should show us all the roles we have.
                show_roles = [x[0] for x in conn.execute_ddl('SHOW CURRENT ROLES')]
                assert steward_role not in show_roles
                assert obj_role1 not in show_roles
                assert obj_role2 not in show_roles
                assert obj_role3 not in show_roles
                assert obj_role4 not in show_roles
                assert 'admin_role' in show_roles

                # SHOW ALL ROLES is only accessible to admins.
                show_roles = [x[0] for x in conn.execute_ddl('SHOW ALL ROLES')]
                assert steward_role in show_roles
                assert obj_role1 in show_roles
                assert obj_role2 in show_roles
                assert obj_role3 in show_roles
                assert obj_role4 in show_roles

                # SHOW ROLE GRANT GROUP on a group the user is not in is
                # only accessible to admins.
                show_roles = [x[0] for x in conn.execute_ddl('SHOW ROLE GRANT GROUP %s' % TEST_USER)]
                assert steward_role in show_roles
                assert obj_role4 in show_roles

    def test_role_priv_show_roles_grant_option(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            steward_role = "role_priv_steward_role"
            obj_role1 = "role_priv_test_role1"
            obj_role2 = "role_priv_test_role2"

            setup_ddls = [
                "DROP ROLE IF EXISTS grantable_role",

                # Setup steward role
                "DROP ROLE IF EXISTS %s" % (steward_role),
                "CREATE ROLE %s" % (steward_role),
                "GRANT ROLE %s TO GROUP %s" % (steward_role, TEST_USER),

                # Setup object roles
                "DROP ROLE IF EXISTS %s" % (obj_role1),
                "CREATE ROLE %s" % (obj_role1),
                "DROP ROLE IF EXISTS %s" % (obj_role2),
                "CREATE ROLE %s" % (obj_role2),

                # Grant specific privilege on specified object to the
                # steward role
                "GRANT SELECT ON TABLE okera_sample.sample TO ROLE %s WITH GRANT OPTION" % (steward_role),
                "GRANT MANAGE_GROUPS ON ROLE %s TO ROLE %s" % (obj_role1, steward_role),
            ]

            for ddl in setup_ddls:
                conn.execute_ddl(ddl)

            print("Running SHOW ROLES tests WITH GRANT OPTION")

            # Switch to the user who has the roles
            ctx.enable_token_auth(token_str=TEST_USER)

            # SHOW ROLES should show us all the roles we have access to.
            show_roles = [x[0] for x in conn.execute_ddl('SHOW ROLES')]
            assert steward_role in show_roles
            assert obj_role1 in show_roles
            assert obj_role2 not in show_roles

            # SHOW ALL ROLES is only accessible to admins or if you had
            # WITH GRANT OPTION
            show_roles = [x[0] for x in conn.execute_ddl('SHOW ALL ROLES')]
            assert steward_role in show_roles
            assert obj_role1 in show_roles
            assert obj_role2 not in show_roles

            # SHOW ROLE GRANT GROUP is only accessible to admins or if you had
            # WITH GRANT OPTION
            show_roles = [x[0] for x in conn.execute_ddl('SHOW ROLE GRANT GROUP %s' % TEST_USER2)]
            assert steward_role not in show_roles
            assert obj_role1 not in show_roles
            assert obj_role2 not in show_roles

    def test_role_priv_show_grant_role(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            steward_role = "role_priv_steward_role"
            obj_role1 = "role_priv_test_role1"
            obj_role2 = "role_priv_test_role2"
            obj_role3 = "role_priv_test_role3"
            obj_role4 = "role_priv_test_role4"

            for level in ['SHOW', 'MANAGE_GROUPS']:
                setup_ddls = [
                    # Setup steward role
                    "DROP ROLE IF EXISTS %s" % (steward_role),
                    "CREATE ROLE %s" % (steward_role),
                    "GRANT ROLE %s TO GROUP %s" % (steward_role, TEST_USER),

                    # Setup object roles
                    "DROP ROLE IF EXISTS %s" % (obj_role1),
                    "CREATE ROLE %s" % (obj_role1),
                    "DROP ROLE IF EXISTS %s" % (obj_role2),
                    "CREATE ROLE %s" % (obj_role2),
                    "DROP ROLE IF EXISTS %s" % (obj_role3),
                    "CREATE ROLE %s" % (obj_role3),
                    "DROP ROLE IF EXISTS %s" % (obj_role4),
                    "CREATE ROLE %s" % (obj_role4),

                    # Grant the last role to us explicitly
                    "GRANT ROLE %s TO GROUP %s" % (obj_role4, TEST_USER),

                    # Grant specific privilege on specified object to the
                    # steward role
                    "GRANT %s ON ROLE %s TO ROLE %s" % (level, obj_role1, steward_role),
                    "GRANT %s ON ROLE %s TO ROLE %s" % (level, obj_role2, steward_role),
                ]

                for ddl in setup_ddls:
                    conn.execute_ddl(ddl)

                print("Running SHOW GRANT ROLE tests for access level: %s" % (level))

                # Switch to the user who has the roles
                ctx.enable_token_auth(token_str=TEST_USER)

                conn.execute_ddl('SHOW GRANT ROLE %s' % obj_role1)

                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                    conn.execute_ddl('SHOW GRANT ROLE %s' % obj_role3)
                self.assertTrue("'roles_test_user' is not an admin" in str(ex_ctx.exception))

                conn.execute_ddl('SHOW GRANT ROLE %s' % obj_role4)

                # Switch to the other user who has no roles
                ctx.enable_token_auth(token_str=TEST_USER2)

                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                    conn.execute_ddl('SHOW GRANT ROLE %s' % obj_role1)
                self.assertTrue("'roles_test_user2' is not an admin" in str(ex_ctx.exception))

                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                    conn.execute_ddl('SHOW GRANT ROLE %s' % obj_role3)
                self.assertTrue("'roles_test_user2' is not an admin" in str(ex_ctx.exception))

                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                    conn.execute_ddl('SHOW GRANT ROLE %s' % obj_role4)
                self.assertTrue("'roles_test_user2' is not an admin" in str(ex_ctx.exception))

                # Switch to admin user
                ctx.disable_auth()

                conn.execute_ddl('SHOW GRANT ROLE %s' % obj_role1)
                conn.execute_ddl('SHOW GRANT ROLE %s' % obj_role3)
                conn.execute_ddl('SHOW GRANT ROLE %s' % obj_role4)

    def test_get_role(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            test_attr_ns = "get_role_ns"
            test_attr = "%s.get_role_attr" % (test_attr_ns)
            test_role1 = "get_role_test_role"
            test_role2 = "get_role_test_role2"

            setup_ddls = [
                # Create an attribute
                "DROP ATTRIBUTE IF EXISTS %s" % (test_attr),
                "CREATE ATTRIBUTE %s" % (test_attr),

                # Setup role 1
                "DROP ROLE IF EXISTS %s" % (test_role1),
                "CREATE ROLE %s" % (test_role1),
                "GRANT ROLE %s TO GROUP %s" % (test_role1, TEST_USER),

                # Add grants of various kinds to ensure they get
                # serialized correctly
                "GRANT VIEW_AUDIT ON SERVER TO ROLE %s" % (test_role1),
                "GRANT ALTER ON DATABASE okera_sample TO ROLE %s" % (test_role1),
                "GRANT SHOW ON TABLE okera_sample.sample TO ROLE %s" % (test_role1),
                "GRANT SELECT(record) ON TABLE okera_sample.sample TO ROLE %s" % (test_role1),
                "GRANT ALL ON ATTRIBUTE NAMESPACE %s TO ROLE %s" % (test_attr_ns, test_role1),
                "GRANT MANAGE_GROUPS ON ROLE okera_public_role TO ROLE %s" % (test_role1),
                "GRANT ALL ON CRAWLER foobar TO ROLE %s" % (test_role1),
                """GRANT SELECT ON DATABASE okera_sample
                   HAVING ATTRIBUTE IN (%s)
                   TRANSFORM %s WITH mask()
                   WHERE record = 'abc'
                   TO ROLE %s""" % (test_attr, test_attr, test_role1),

                # Setup role 2, don't grant to user
                "DROP ROLE IF EXISTS %s" % (test_role2),
                "CREATE ROLE %s" % (test_role2),
            ]

            for ddl in setup_ddls:
                conn.execute_ddl(ddl)

            # As admin, ensure we have access to the two roles we created
            role1 = get_role(conn, test_role1).role
            assert role1 is not None
            assert len(role1.privileges) == 8
            # Admin has 'ALL' access on any role so access_level should contain 'ALL'
            assert TAccessPermissionLevel.ALL in role1.access_levels

            role2 = get_role(conn, test_role2).role
            assert role2 is not None
            assert len(role2.privileges) == 0
            # Admin has 'ALL' access on any role so access_level should contain 'ALL'
            assert TAccessPermissionLevel.ALL in role1.access_levels

            # Switch to the user who has the roles
            ctx.enable_token_auth(token_str=TEST_USER)

            expected_privileges = [
                (TAccessPermissionScope.DATABASE, TAccessPermissionLevel.ALTER, 'okera_sample', '', '', None, None, None),
                (TAccessPermissionScope.TABLE, TAccessPermissionLevel.SHOW, 'okera_sample', 'sample', '', None, None, None),
                (TAccessPermissionScope.SERVER, TAccessPermissionLevel.VIEW_AUDIT, '', '', '', None, None, None),
                (TAccessPermissionScope.CRAWLER, TAccessPermissionLevel.ALL, None, None, None, 'foobar', None, None),
                (TAccessPermissionScope.COLUMN, TAccessPermissionLevel.SELECT, 'okera_sample', 'sample', 'record', None, None, None),
                (TAccessPermissionScope.ROLE, TAccessPermissionLevel.MANAGE_GROUPS, '', '', '', None, 'okera_public_role', None),
                (TAccessPermissionScope.ATTRIBUTE_NAMESPACE, TAccessPermissionLevel.ALL, '', '', '', None, None, test_attr_ns),
                (TAccessPermissionScope.DATABASE, TAccessPermissionLevel.SELECT, 'okera_sample', None, None, None, None, None),
            ]

            role1 = get_role(conn, test_role1).role
            assert role1 is not None
            assert len(role1.privileges) == 8
            # TEST_USER would not have any access on it's own role unless specified explicitly
            # access_levels should be an empty list.
            assert not role1.access_levels

            role2 = get_role(conn, "okera_public_role").role
            assert role2 is not None
            # TEST_USER has 'MANAGE_GROUPS' privilege on 'okera_public_role'
            assert TAccessPermissionLevel.MANAGE_GROUPS in role2.access_levels

            found_abac_priv = False
            for priv in role1.privileges:
                priv_tuple = (
                    priv.scope, priv.privilege_level, priv.db_name,
                    priv.table_name, priv.column_name, priv.crawler,
                    priv.role, priv.attribute_namespace)
                assert priv_tuple in expected_privileges

                if priv.transform_expression:
                    found_abac_priv = True
                    assert test_attr_ns in priv.transform_expression
                    assert test_attr_ns in priv.attribute_expression
                    assert 'record' in priv.filter
            assert found_abac_priv

            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                get_role(conn, test_role2)
            self.assertTrue("Role 'get_role_test_role2' does not exist" in str(ex_ctx.exception))

    def test_list_roles_non_paginated(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            steward_role = "list_roles_steward_role"
            test_attr_ns = "get_role_ns"
            test_attr = "%s.get_role_attr" % (test_attr_ns)
            test_role = "list_roles_role1"
            test_role2 = "manageable_test_role"
            test_role3 = "manage_groups_test_role"
            test_group = "list_roles_group1"
            test_db = "list_roles_db1"
            test_tbl = "list_roles_tbl1"

            setup_ddls = [
                # Create an attribute
                "DROP ATTRIBUTE IF EXISTS %s" % (test_attr),
                "CREATE ATTRIBUTE %s" % (test_attr),

                # Setup DBs
                "DROP DATABASE IF EXISTS %s CASCADE" % (test_db),
                "CREATE DATABASE %s" % (test_db),
                "CREATE TABLE %s.%s (c1 int)" % (test_db, test_tbl),

                # Setup role
                "DROP ROLE IF EXISTS %s" % (test_role),
                "CREATE ROLE %s" % (test_role),
                "DROP ROLE IF EXISTS %s" % (test_role2),
                "CREATE ROLE %s" % (test_role2),
                "DROP ROLE IF EXISTS %s" % (test_role3),
                "CREATE ROLE %s" % (test_role3),
                """GRANT SELECT ON TABLE %s.%s
                   HAVING ATTRIBUTE IN (%s)
                   TO ROLE %s""" % (test_db, test_tbl, test_attr, test_role),
                "GRANT ROLE %s TO GROUP %s" % (test_role, test_group),

                "DROP ROLE IF EXISTS %s" % (steward_role),
                "CREATE ROLE %s" % (steward_role),
                "GRANT ALL ON ROLE %s TO ROLE %s" % (test_role, steward_role),
                "GRANT SHOW ON ROLE %s TO ROLE %s" % (test_role2, steward_role),
                "GRANT MANAGE_GROUPS ON ROLE %s TO ROLE %s" % (test_role3, steward_role),
                "GRANT ROLE %s TO GROUP %s" % (steward_role, TEST_USER),
            ]

            ctx.disable_auth()

            for ddl in setup_ddls:
                conn.execute_ddl(ddl)

            for (user, has_access) in [(None, True), (TEST_USER, True), (TEST_USER2, False)]:
                if user:
                    print("Running list_roles tests with user: %s" % user)
                    ctx.enable_token_auth(token_str=user)
                else:
                    print("Running list_roles tests as admin")
                    ctx.disable_auth()

                # All roles
                params = _thrift_api.TListRolesParams()
                roles = list_roles(conn, params).roles
                if has_access:
                    assert contains_role(test_role, roles)
                else:
                    assert not contains_role(test_role, roles)

                # Roles with name substring match
                params = _thrift_api.TListRolesParams()
                params.role_name_substring = test_role[1:-1]
                roles = list_roles(conn, params).roles
                if has_access:
                    assert len(roles) == 1
                    assert contains_role(test_role, roles)
                else:
                    assert len(roles) == 0

                params = _thrift_api.TListRolesParams()
                params.role_name_substring = test_role[1:-1] + "not_exist"
                roles = list_roles(conn, params).roles
                assert len(roles) == 0

                # Roles with group substring match
                params = _thrift_api.TListRolesParams()
                params.group_name_substring = test_group[1:-1]
                roles = list_roles(conn, params).roles
                if has_access:
                    assert len(roles) == 1
                    assert contains_role(test_role, roles)
                else:
                    assert len(roles) == 0

                params = _thrift_api.TListRolesParams()
                params.group_name_substring = test_group[1:-1] + "not_exist"
                roles = list_roles(conn, params).roles
                assert len(roles) == 0

                # Roles with target user
                params = _thrift_api.TListRolesParams()
                params.target_user = test_group
                roles = list_roles(conn, params).roles
                if has_access:
                    assert contains_role(test_role, roles)
                else:
                    assert not contains_role(test_role, roles)

                params = _thrift_api.TListRolesParams()
                params.target_user = test_group + "not_exist"
                roles = list_roles(conn, params).roles
                assert not contains_role(test_role, roles)

                # Roles with catalog object filter
                params = _thrift_api.TListRolesParams()
                params.catalog_object_names = ["%s.%s" % (test_db, test_tbl)]
                roles = list_roles(conn, params).roles
                if has_access:
                    assert contains_role(test_role, roles)
                else:
                    assert not contains_role(test_role, roles)

                params = _thrift_api.TListRolesParams()
                params.catalog_object_names = ["%s.%s_not_exist" % (test_db, test_tbl)]
                roles = list_roles(conn, params).roles
                assert not contains_role(test_role, roles)

                params = _thrift_api.TListRolesParams()
                params.catalog_object_names = ["%s" % (test_db)]
                roles = list_roles(conn, params).roles
                if has_access:
                    assert contains_role(test_role, roles)
                else:
                    assert not contains_role(test_role, roles)

                params = _thrift_api.TListRolesParams()
                params.catalog_object_names = ["%s_not_exist" % (test_db)]
                roles = list_roles(conn, params).roles
                assert not contains_role(test_role, roles)

                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                    params = _thrift_api.TListRolesParams()
                    params.catalog_object_names = ["db1", "db2"]
                    list_roles(conn, params).roles
                self.assertTrue("More than one catalog object filter is not supported" in str(ex_ctx.exception))

                # Roles with scope filter
                params = _thrift_api.TListRolesParams()
                params.scopes = [TAccessPermissionScope.TABLE]
                roles = list_roles(conn, params).roles
                if has_access:
                    assert contains_role(test_role, roles)
                else:
                    assert not contains_role(test_role, roles)

                params = _thrift_api.TListRolesParams()
                params.scopes = [TAccessPermissionScope.DATABASE]
                roles = list_roles(conn, params).roles
                assert not contains_role(test_role, roles)

                params = _thrift_api.TListRolesParams()
                params.scopes = [TAccessPermissionScope.DATABASE, TAccessPermissionScope.TABLE]
                roles = list_roles(conn, params).roles
                if has_access:
                    assert contains_role(test_role, roles)
                else:
                    assert not contains_role(test_role, roles)

                # Check scope filter + catalog object filter is not allowed
                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                    params = _thrift_api.TListRolesParams()
                    params.catalog_object_names = ["db1", "db2"]
                    params.scopes = [TAccessPermissionScope.DATABASE]
                    list_roles(conn, params).roles
                self.assertTrue("both an access scope filter and a catalog object filter" in str(ex_ctx.exception))

                # Check manageableOnly filter
                params = _thrift_api.TListRolesParams()
                params.manageable_only = True

                roles = list_roles(conn, params).roles
                if has_access:
                    if user:
                        # Steward should only be able to see roles
                        # that they have privilege on, ie its own roles
                        # should be filtered out
                        assert not contains_role(steward_role, roles)
                    assert contains_role(test_role, roles)
                else:
                    assert not contains_role(test_role, roles)

                # Check levels filter
                # This filter works with effective permissions, ie
                # ALL will always be added to the specified level, since
                # ALL implies all other levels.
                params = _thrift_api.TListRolesParams()
                params.levels = [TAccessPermissionLevel.MANAGE_GROUPS]

                roles = list_roles(conn, params).roles
                if has_access:
                    # both steward and admin should see the role with MANAGE_GRANTS
                    assert contains_role(test_role3, roles)
                    if user:
                        # steward should not see the role with SHOW
                        assert not contains_role(test_role2, roles)
                else:
                    assert not contains_role(test_role2, roles)
                    assert not contains_role(test_role3, roles)

    def test_list_roles_paginated(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            steward_role = "list_roles_steward_role"

            setup_ddls = [
                "DROP ROLE IF EXISTS %s" % (steward_role),
                "CREATE ROLE %s" % (steward_role),
                "GRANT ROLE %s TO GROUP %s" % (steward_role, TEST_USER)
            ]

            expected_order = []

            for idx in range(9):
                role_name = "list_roles_paginated_role_%s" % idx
                setup_ddls += [
                    "DROP ROLE IF EXISTS %s" % (role_name),
                    "CREATE ROLE %s" % (role_name),
                    "GRANT ALL ON ROLE %s TO ROLE %s" % (role_name, steward_role)
                ]

                expected_order.append(role_name)

            ctx.disable_auth()

            for ddl in setup_ddls:
                conn.execute_ddl(ddl)

            ctx.enable_token_auth(token_str=TEST_USER)

            # Paginate filtered roles
            prev_tokens = []

            params = _thrift_api.TListRolesParams()
            params.role_name_substring = "roles_paginated_role"
            idx = 0
            while True:
                res = list_roles(conn, params, limit=2)
                assert len(res.roles) <= 2
                assert res.total_count == 9

                # For every non-None prev token, we will record
                # it so we can walk it backwards and ensure it works.
                if res.prev_page_token is not None:
                    prev_tokens.append(res.prev_page_token)

                for role in res.roles:
                    assert role.role_name == expected_order[idx]
                    idx += 1


                if res.next_page_token:
                    params.page_token = res.next_page_token
                else:
                    break

            # Iterate over the prev tokens to ensure we get back
            # the same values.
            idx = 0
            for prev_token in prev_tokens:
                params.page_token = prev_token
                res = list_roles(conn, params, limit=2)
                assert len(res.roles) == 2
                assert res.total_count == 9
                for role in res.roles:
                    assert role.role_name == expected_order[idx]
                    idx += 1

    def test_list_roles_access_levels(self):
        ROLE1 = "test_list_roles_access_levels_role1"
        USER1 = "test_list_roles_access_levels_user1"

        ROLE2 = "test_list_roles_access_levels_role2"
        USER2 = "test_list_roles_access_levels_user2"

        ROLE3 = "test_list_roles_access_levels_role3"
        USER3 = "test_list_roles_access_levels_user3"

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:

            ddls = [
                "DROP ROLE IF EXISTS %s" % (ROLE1),
                "CREATE ROLE %s" % (ROLE1),
                "GRANT ROLE %s TO GROUP %s" % (ROLE1, USER1),

                "DROP ROLE IF EXISTS %s" % (ROLE2),
                "CREATE ROLE %s" % (ROLE2),
                "GRANT ROLE %s TO GROUP %s" % (ROLE2, USER2),

                "DROP ROLE IF EXISTS %s" % (ROLE3),
                "CREATE ROLE %s" % (ROLE3),
                "GRANT ROLE %s TO GROUP %s" % (ROLE3, USER3),

                # Grant privileges to ROLE1 on ROLE2
                "GRANT ALL ON ROLE %s TO ROLE %s" % (ROLE2, ROLE1),
                "GRANT SHOW ON ROLE %s TO ROLE %s" % (ROLE2, ROLE1),
                "GRANT DROP ON ROLE %s TO ROLE %s" % (ROLE2, ROLE1),
                "GRANT ALTER ON ROLE %s TO ROLE %s" % (ROLE2, ROLE1),
                "GRANT MANAGE_GROUPS ON ROLE %s TO ROLE %s" % (ROLE2, ROLE1),

                # Grant privileges to ROLE1 on ROLE2
                "GRANT SHOW ON ROLE %s TO ROLE %s" % (ROLE1, ROLE2),
                "GRANT DROP ON ROLE %s TO ROLE %s" % (ROLE1, ROLE2),

                # No privileges for ROLE3
            ]

            for ddl in ddls:
                conn.execute_ddl(ddl)

            # -------- list roles for admin  -----------
            print("Running access_levels test : list_roles for %s" % ('admin'))
            # If authorization is disabled or if user is an admin or if it's an internal
            # request, the user will always have 'ALL' permission on any role
            ctx.disable_auth()

            params = _thrift_api.TListRolesParams()
            roles = list_roles(conn, params).roles

            # admin would have access to all the roles. we are verifying just below four.
            assert contains_role(OKERA_PUBLIC_ROLE, roles)
            assert contains_role(ROLE1, roles)
            assert contains_role(ROLE2, roles)
            assert contains_role(ROLE3, roles)

            for role in roles:
                if role.role_name in [ROLE1, ROLE2, ROLE3, OKERA_PUBLIC_ROLE]:
                    assert TAccessPermissionLevel.ALL in role.access_levels

            # -------- list roles for USER1 -----------
            print("Running access_levels test : list_roles for %s" % (USER1))
            ctx.enable_token_auth(token_str=USER1)

            params = _thrift_api.TListRolesParams()
            roles = list_roles(conn, params).roles

            assert contains_role(OKERA_PUBLIC_ROLE, roles)
            assert contains_role(ROLE1, roles)
            assert contains_role(ROLE2, roles)
            assert not contains_role(ROLE3, roles)

            for role in roles:
                if role.role_name == OKERA_PUBLIC_ROLE:
                    # USER1 does not have any explicit access on OKERA_PUBLIC_ROLE
                    # access_levels should be an empty list
                    assert not role.access_levels
                if role.role_name == ROLE1:
                    # USER1 would not have any access on it's own role unless specified explicitly
                    # access_levels should be an empty list
                    assert not role.access_levels
                if role.role_name == ROLE2:
                    # privileges that USER1 has on ROLE2
                    assert TAccessPermissionLevel.ALL in role.access_levels
                    assert TAccessPermissionLevel.SHOW in role.access_levels
                    assert TAccessPermissionLevel.DROP in role.access_levels
                    assert TAccessPermissionLevel.ALTER in role.access_levels
                    assert TAccessPermissionLevel.MANAGE_GROUPS in role.access_levels

            # -------- list roles for USER2 -----------
            print("Running access_levels test : list_roles for %s" % (USER2))
            ctx.enable_token_auth(token_str=USER2)

            params = _thrift_api.TListRolesParams()
            roles = list_roles(conn, params).roles

            assert contains_role(OKERA_PUBLIC_ROLE, roles)
            assert contains_role(ROLE1, roles)
            assert contains_role(ROLE2, roles)
            assert not contains_role(ROLE3, roles)

            for role in roles:
                if role.role_name == OKERA_PUBLIC_ROLE:
                    # USER2 does not have any explicit access on OKERA_PUBLIC_ROLE
                    # access_levels should be an empty list
                    assert not role.access_levels
                if role.role_name == ROLE1:
                    # privileges that USER2 has on ROLE1
                    assert TAccessPermissionLevel.ALL not in role.access_levels
                    assert TAccessPermissionLevel.SHOW in role.access_levels
                    assert TAccessPermissionLevel.DROP in role.access_levels
                    assert TAccessPermissionLevel.ALTER not in role.access_levels
                    assert TAccessPermissionLevel.MANAGE_GROUPS not in role.access_levels
                if role.role_name == ROLE2:
                    # USER2 would not have any access on it's own role unless specified explicitly
                    # access_levels should be an empty list
                    assert not role.access_levels

            # -------- list roles for USER3 -----------
            print("Running access_levels test : list_roles for %s" % (USER3))
            ctx.enable_token_auth(token_str=USER3)

            params = _thrift_api.TListRolesParams()
            roles = list_roles(conn, params).roles

            assert contains_role(OKERA_PUBLIC_ROLE, roles)
            assert not contains_role(ROLE1, roles)
            assert not contains_role(ROLE2, roles)
            assert contains_role(ROLE3, roles)

            for role in roles:
                if role.role_name == OKERA_PUBLIC_ROLE:
                    # USER3 does not have any explicit access on OKERA_PUBLIC_ROLE
                    # access_levels should be an empty list
                    assert not role.access_levels
                if role.role_name == ROLE3:
                    # USER3 would not have any access on it's own role unless specified explicitly
                    # access_levels should be an empty list
                    assert not role.access_levels

    def test_role_list_roles_user_role(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            steward_role = "role_priv_steward_user_role"
            as_owner_role = "_okera_internal_role_%s" % (TEST_USER)
            test_db = "user_role_test_db"

            ctx.disable_auth()
            setup_ddls = [
                "DROP DATABASE IF EXISTS %s" % (test_db),

                # Setup steward role
                "DROP ROLE IF EXISTS %s" % (as_owner_role),
                "DROP ROLE IF EXISTS %s" % (steward_role),
                "CREATE ROLE %s" % (steward_role),
                "GRANT ROLE %s TO GROUP %s" % (steward_role, TEST_USER),

                # Grant specific privilege on specified object to the
                # steward role
                "GRANT CREATE_AS_OWNER ON CATALOG TO ROLE %s" % (steward_role),
            ]

            for ddl in setup_ddls:
                conn.execute_ddl(ddl)

            # Switch to the user
            ctx.enable_token_auth(token_str=TEST_USER)

            params = _thrift_api.TListRolesParams()
            roles = list_roles(conn, params).roles
            assert len(roles) >= 2
            assert contains_role(OKERA_PUBLIC_ROLE, roles)
            assert contains_role(steward_role, roles)
            assert not contains_role(as_owner_role, roles)

            conn.execute_ddl("CREATE DATABASE %s" % test_db)

            params = _thrift_api.TListRolesParams()
            roles = list_roles(conn, params).roles
            assert len(roles) >= 3
            assert contains_role(OKERA_PUBLIC_ROLE, roles)
            assert contains_role(steward_role, roles)
            assert contains_role(as_owner_role, roles)

            assert get_role_by_name(steward_role, roles).is_user_role == False
            assert get_role_by_name(as_owner_role, roles).is_user_role == True

            assert get_role(conn, as_owner_role).role.is_user_role == True
            assert get_role(conn, steward_role).role.is_user_role == False

if __name__ == "__main__":
    unittest.main()



# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests for json schema discovery
#
# pylint: disable=line-too-long
# pylint: disable=too-many-locals
# pylint: disable=duplicate-code

from datetime import datetime, timedelta
import decimal
from urllib.parse import urlparse
import json
import os
import time
import unittest

import boto3
import requests

from okera import _thrift_api
from okera.tests import pycerebro_test_common as common
from okera.tests import file_format_test_common as fileFormat
from okera._thrift_api import TTypeId

TEST_DB = 'json_test_db'
TEST_TABLE = TEST_DB + ".tbl"
DEFAULT_ENABLE_PRESTO = False

S3_TEST_FILES = [
    's3://cerebrodata-test/chase/json-empty/chase-json-test.json',
    's3://cerebrodata-test/chase/json-simple/chase-json-test.json',
    's3://cerebrodata-dev/json-random-testing/cases/case1/data.json',
    's3://cerebrodata-dev/json-random-testing/cases/case1/data2.json',
    's3://cerebrodata-dev/json-random-testing/cases/case1/data3.json',
    's3://cerebrodata-dev/json-random-testing/cases/case1/data4.json',
    's3://cerebrodata-dev/json-random-testing/cases/case1/data5.json',
    's3://cerebrodata-dev/json-random-testing/cases/len-mismatch/data.json',
    's3://cerebrodata-dev/json-random-testing/cases/merge-basic/data.json',
    's3://cerebrodata-dev/json-random-testing/cases/merge2/data.json',
    's3://cerebrodata-dev/json-random-testing/cases/merge3/data.json',
    's3://cerebrodata-dev/json-random-testing/cases/merge4/data.json',

    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1211.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1211_fees.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1216.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1216_with_subscriptionLimit.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1238.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1238_1.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1238_2.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1238_3.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1238_4.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1238_5.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1238_6.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1238_7.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1238_8.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1238_9.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1238_parquet_ledger_balance.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1238_parquet_ledger_transaction.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1238_parquet_party.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1238_parquet_product.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1238_parquet_subscription.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1269.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1269_avro2.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1269_avro_null_fields.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1269_partitioned-files.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1269_problem_file.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1269_success1.json',
    's3://cerebro-customers/chase/jsonified/downloads_chase_zd1269_success2.json',

    # simdjson website example jsons
    's3://cerebrodata-dev/simdjson/jsonexamples/small/jsoniter_scala/che-1.geo.json',
    's3://cerebrodata-dev/simdjson/jsonexamples/small/jsoniter_scala/che-2.geo.json',
    's3://cerebrodata-dev/simdjson/jsonexamples/small/jsoniter_scala/che-3.geo.json',
    's3://cerebrodata-dev/simdjson/jsonexamples/small/smalldemo.json',
    's3://cerebrodata-dev/simdjson/jsonexamples/google_maps_api_compact_response.json',
    's3://cerebrodata-dev/simdjson/jsonexamples/mesh.json',
]

S3_TEST_ERROR_FILES = [
    # Fails to infer the schema, top level is array, not record which is not valid.
    's3://cerebrodata-dev/simdjson/jsonexamples/small/truenull.json',
    's3://cerebrodata-dev/simdjson/jsonexamples/amazon_cellphones.ndjson',

    # Fails with json.decoder.JSONDecodeError: Expecting property name enclosed
    # in double quotes: line 1 column 2 (char 1)
    's3://cerebrodata-dev/simdjson/jsonexamples/small/adversarial.json',
    's3://cerebrodata-dev/simdjson/jsonexamples/small/demo.json',

    # Column name in file has a quote in it, which is not a valid column name
    's3://cerebrodata-dev/simdjson/jsonexamples/small/flatadversarial.json',

    # Longer than 500K in one record
    's3://cerebrodata-dev/simdjson/jsonexamples/marine_ik.json',
    's3://cerebrodata-dev/simdjson/jsonexamples/update-center.json',
]

# URLs that return json. The second argument is if the JSON has a root level key
# that is really the entire table
TEST_URLS = [
    # These work. Taken from: https://github.com/jdorfman/awesome-json-datasets
    ('https://blockchain.info/unconfirmed-transactions?format=json', 'txs'),
    ('https://data.police.uk/api/leicestershire/neighbourhoods', ''),
    ('https://data.police.uk/api/forces', ''),
    ('https://pokeapi.co/api/v2/pokemon/1/', 'abilities'),
    ('https://pokeapi.co/api/v2/ability/1', 'effect_changes'),
    ('https://data.parliament.scot/api/members', ''),
    ('https://earthquake.usgs.gov/earthquakes/feed/v1.0/summary/all_hour.geojson', 'features'),
    ('http://showtimes.everyday.in.th/api/v2/movie/', None),
    ('http://showtimes.everyday.in.th/api/v2/theater/', None),
    ('https://www.govtrack.us/api/v2/role?current=true&role_type=representative&limit=438', 'objects'),
    ('https://data.nasa.gov/resource/2vr3-k9wn.json', ''),
    ('http://api.nobelprize.org/v1/country.json', 'countries'),
    ('http://api.nobelprize.org/v1/laureate.json', 'laureates'),
    ('http://api.nobelprize.org/v1/prize.json', 'prizes'),
    ('https://raw.githubusercontent.com/Biuni/PokemonGO-Pokedex/master/pokedex.json', 'pokemon'),
    ('https://raw.githubusercontent.com/prust/wikipedia-movie-data/master/movies.json', ''),

    # simdjson test cases
    ('https://raw.githubusercontent.com/simdjson/simdjson/master/jsonexamples/apache_builds.json', 'jobs'),
    ('https://raw.githubusercontent.com/simdjson/simdjson/master/jsonexamples/github_events.json', ''),
    ('https://raw.githubusercontent.com/simdjson/simdjson/master/jsonexamples/google_maps_api_compact_response.json', None),
    ('https://raw.githubusercontent.com/simdjson/simdjson/master/jsonexamples/google_maps_api_response.json', ''),
    ('https://raw.githubusercontent.com/simdjson/simdjson/master/jsonexamples/instruments.json', ''),
    ('https://raw.githubusercontent.com/simdjson/simdjson/master/jsonexamples/mesh.json', None),
    ('https://raw.githubusercontent.com/simdjson/simdjson/master/jsonexamples/mesh.pretty.json', ''),
    ('https://raw.githubusercontent.com/simdjson/simdjson/master/jsonexamples/random.json', ''),
    ('https://raw.githubusercontent.com/simdjson/simdjson/master/jsonexamples/repeat.json', ''),
    ('https://raw.githubusercontent.com/simdjson/simdjson/master/jsonexamples/repeat.json', 'result'),
    ('https://raw.githubusercontent.com/simdjson/simdjson/master/jsonexamples/tree-pretty.json', ''),
    ('https://raw.githubusercontent.com/simdjson/simdjson/master/jsonexamples/twitter.json', 'statuses'),
    ('https://raw.githubusercontent.com/simdjson/simdjson/master/jsonexamples/twitter_api_compact_response.json', ''),
    ('https://raw.githubusercontent.com/simdjson/simdjson/master/jsonexamples/twitter_timeline.json', ''),
    ('https://raw.githubusercontent.com/simdjson/simdjson/master/jsonexamples/twitterescaped.json', 'statuses'),
]

TEST_ERROR_URLS = [
    # These work but the contents of these change constantly
    #('https://blockchain.info/latestblock', None),
    #('http://api.open-notify.org/iss-now.json', None),
    #('http://api.open-notify.org/astros.json', None),
    #('https://api.exchangerate-api.com/v4/latest/USD', None),
    #('https://api.exchangerate-api.com/v4/latest/GBP', None),
    #('https://data.gov.au/geoserver/abc-local-stations/wfs?request=GetFeature&typeName=ckan_d534c0e9_a9bf_487b_ac8f_b7877a09d162&outputFormat=json', None),

    # FIXME: Keys normalize to _ instead of -
    #('https://data.police.uk/api/crimes-street-dates', ''),
    #('https://www.govtrack.us/api/v2/role?current=true&role_type=senator', 'objects'),

    # FIXME: Keys normalize @ to empty so column is missing
    #('https://data.nasa.gov/resource/y77d-th95.json', ''),

    # FIXME: Keys have fields with spaces in them
    #('https://api.github.com/gists', ''),

    # TODO: keys are ints which is generating sometihing invalid?
    #('https://www.ncdc.noaa.gov/cag/global/time-series/globe/land_ocean/ytd/12/1880-2016.json', None),

    # TODO: schema length over 4000
    #('https://api.github.com/events', ''),

    # The top level is an array (not record, so this is an invalid json)
    #('https://raw.githubusercontent.com/simdjson/simdjson/master/jsonexamples/marine_ik.json', ''),
    # This contains a single record over 500K
    #('https://raw.githubusercontent.com/simdjson/simdjson/master/jsonexamples/numbers.json', ''),
]

TEST_DB = 'json_test_db'
TEST_TABLE = TEST_DB + ".tbl"
S3 = boto3.resource('s3')
TEST_BUCKET = 'cerebrodata-dev'
TEST_DIR_PREFIX = 'json-random-testing'
FILE_FORMAT = 'JSON'

class JsonIntegrationTest(fileFormat.FileFormatTestBase):

    def __test_s3(self, path, allow_missing=False, test_scan_api=False):
        # Read it directly using python libraries
        o = urlparse(path, allow_fragments=False)
        obj = S3.Object(o.netloc, o.path.lstrip('/'))
        expected_json = []
        for line in obj.get()['Body'].read().decode('utf-8').strip().split('\n'):
            expected_json.append(json.loads(line))
        self._test_path(path, expected_json, TEST_DB, TEST_TABLE, allow_missing,
                        test_scan_api=test_scan_api, file_format=FILE_FORMAT)

    def __test_url(self, url, key, expected_json=None):
        print("Testing (%s, %s)" % (url, key))
        expected_json = json.loads(requests.get(url).text)
        if key is None:
            expected_json = [expected_json]
            self._test_path(url, expected_json, TEST_DB, TEST_TABLE, allow_missing=True,
                            file_format=FILE_FORMAT)
            return

        # There is a key, we need to transform the data
        if key:
            expected_json = expected_json[key]
        elif isinstance(expected_json, dict):
            expected_json = [expected_json]

        data = []
        for record in expected_json:
            data.append(json.dumps(record))
        path = self._write_s3(TEST_DIR_PREFIX, 'json-http-testing/data.json',
                              '\n'.join(data))
        self._test_path(path, expected_json, TEST_DB, TEST_TABLE, allow_missing=True,
                        file_format=FILE_FORMAT)

    def _run_random_tests(self, generator, iters, path):
        # Run random generator tests
        if iters is None:
            iters = 100

        for idx in range(0, iters):
            generator.new_schema()
            data = generator.generate()
            print("Generated json for test case%s:\n%s" % \
                  (idx, json.dumps(data, indent=2, sort_keys=True)))
            filename = '%s/test-%s.json' % (path, idx)
            output = []
            for record in data:
                output.append(json.dumps(record))
            full_path = self._write_s3(TEST_DIR_PREFIX, filename, '\n'.join(output))
            self._test_path(full_path, data, TEST_DB, TEST_TABLE, allow_missing=True,
                            file_format=FILE_FORMAT)

    @unittest.skip("Example for use during development to test a specific dataset.")
    def test_single_dataset(self):
        self.__test_s3('s3://cerebrodata-test/chase/json-simple/chase-json-test.json')

    @unittest.skip("Example for use during development to test a specific dataset.")
    def test_single_url(self):
        self.__test_url('https://blockchain.info/latestblock', None)

    def test_s3(self):
        for f in S3_TEST_FILES:
            self.__test_s3(f, allow_missing=True, test_scan_api=True)

    def valid_json_api(self, url):
        ctx = common.get_test_context()
        disable_prop = "WITH SERDEPROPERTIES('new.line.json'='true') "
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('CREATE DATABASE IF NOT EXISTS %s' % TEST_DB)

            # Explicit schema
            conn.execute_ddl('DROP TABLE IF EXISTS %s' % TEST_TABLE)
            conn.execute_ddl(
                ("CREATE EXTERNAL TABLE %s(revenue BIGINT) STORED AS JSON " +
                 "LOCATION '%s'") % (TEST_TABLE, url))
            json = conn.scan_as_json(TEST_TABLE, max_records=2)
            # TODO: not sure how stable this live endpoint is so making the validation
            # less strict
            self.assertEqual(2, len(json))
            self.assertEqual(1, len(json[0]))
            self.assertTrue('revenue' in json[0])

            # Explicitly disable new line json, this won't scan right
            conn.execute_ddl('DROP TABLE %s' % TEST_TABLE)
            conn.execute_ddl(
                ("CREATE EXTERNAL TABLE %s(revenue BIGINT) %s " +
                 "STORED AS JSON " +
                 "LOCATION '%s'") % (TEST_TABLE, disable_prop, url))
            json = conn.scan_as_json(TEST_TABLE, max_records=2)
            self.assertEqual([], json)

            # Infer full schema
            conn.execute_ddl('DROP TABLE %s' % TEST_TABLE)
            conn.execute_ddl(
                ("CREATE EXTERNAL TABLE %s " +
                 "LIKE JSON '%s' " +
                 "STORED AS JSON " +
                 "LOCATION '%s'") % (TEST_TABLE, url, url))
            json = conn.scan_as_json(TEST_TABLE, max_records=2)
            self.assertEqual(2, len(json))
            self.assertTrue(len(json[0]) > 20)
            self.assertTrue('revenue' in json[0])
            self.assertTrue('symbol' in json[0])

            # Disable new line json, we won't infer it right
            conn.execute_ddl('DROP TABLE %s' % TEST_TABLE)
            with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                conn.execute_ddl(
                    ("CREATE EXTERNAL TABLE %s " +
                     "LIKE JSON '%s' %s " +
                     "STORED AS JSON " +
                     "LOCATION '%s'") % (TEST_TABLE, url, disable_prop, url))
            print(ex_ctx.exception)
            self.assertTrue('Failed to read JSON schema' in str(ex_ctx.exception))

    def test_json_api(self):
        url = 'https://financialmodelingprep.com/api/v3/income-statement/AAPL?' +\
            'limit=10&apikey='
        # Hardcoded apikey
        self.valid_json_api(url + 'demo')
        # Dynamically returned api key
        self.valid_json_api(url + '${script:/get-financialmodelingprep-apikey.sh}')

    def test_json_api_path(self):
        ctx = common.get_test_context()
        path_prop = "WITH SERDEPROPERTIES('json.records-start.path'='content') "
        url = 'https://gist.githubusercontent.com/nongli/' +\
            'd7e30635e466a9352bbcaacfb36a1c09/raw/' +\
            'fea92d963c724fd2d6e1e71fd2c9cba7daa3e277/gistfile1.txt'

        with common.get_planner(ctx) as conn:
            conn.execute_ddl('CREATE DATABASE IF NOT EXISTS %s' % TEST_DB)

            conn.execute_ddl('DROP TABLE IF EXISTS %s' % TEST_TABLE)
            conn.execute_ddl(
                ("CREATE EXTERNAL TABLE %s " +
                 "LIKE JSON '%s' %s" +
                 "STORED AS JSON " +
                 "LOCATION '%s'") % (TEST_TABLE, url, path_prop, url))
            json = conn.scan_as_json(TEST_TABLE)
            self.assertEqual(1, len(json))
            self.assertEqual(len(json[0]), 5)
            self.assertEqual(json[0]['bdrid'], '3095')
            self.assertEqual(json[0]['subtype'], 'OFFICE')

            # Set the config on the DB, it should apply to all tables in it
            conn.execute_ddl('DROP DATABASE %s CASCADE' % TEST_DB)
            conn.execute_ddl('CREATE DATABASE %s' % TEST_DB)
            conn.execute_ddl(
                ("ALTER DATABASE %s SET DBPROPERTIES(" + \
                "'json.records-start.path'='content')") % TEST_DB)
            conn.execute_ddl(
                ("CREATE EXTERNAL TABLE %s " +
                 "LIKE JSON '%s'" +
                 "STORED AS JSON " +
                 "LOCATION '%s'") % (TEST_TABLE, url, url))
            json = conn.scan_as_json(TEST_TABLE)
            self.assertEqual(1, len(json))
            self.assertEqual(len(json[0]), 5)
            self.assertEqual(json[0]['bdrid'], '3095')
            self.assertEqual(json[0]['subtype'], 'OFFICE')

    def test_json_count(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            data = conn.scan_as_json(
                'select count(*) as cnt from rs_json_format.airlines_mini',
                strings_as_utf8=False,
                min_task_size=5 * 1000)
            # for this dataset we expect exactly 1 task splits as json is not splittable
            self.assertTrue(len(data) == 1)
            sum = 0
            for cnt in data:
                sum += cnt['cnt']
            self.assertEqual(sum, 100)

    @unittest.skipIf(common.TEST_LEVEL != "all", "This writes to the same location in s3")
    def test_type_conversion(self):
        # This test does not use schema infernce and instead explicitly creates the
        # types with some interesting values
        data = {
            'c1': 0,
            'c2': None,
            'c3': '',
            'c4': '123',
            'c5': '4.56',
            'c6': 'value',
            'c7': 'vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv',
            'c8': '2020-01-01',
            'c9': '2020-01-01 01:01:01',
        }

        types = [
            ('STRING', None),
            ('BOOLEAN', bool),
            ('TINYINT', int),
            ('SMALLINT', int),
            ('INT', int),
            ('BIGINT', int),
            ('FLOAT', float),
            ('DOUBLE', float),
            ('DATE', None),
            ('TIMESTAMP', None),

            ('VARCHAR(1)', None),
            ('VARCHAR(5)', None),
            ('VARCHAR(40)', None),

            ('CHAR(1)', None),
            ('CHAR(5)', None),
            ('CHAR(40)', None),

            ('DECIMAL(1, 0)', None),
            ('DECIMAL(10, 1)', decimal.Decimal),
            ('DECIMAL(19, 2)', decimal.Decimal),
            ('DECIMAL(38, 10)', decimal.Decimal),
        ]

        for t in types:
            cols = []
            for i in range(1, len(data) + 1):
                cols.append('c' + str(i) + ' ' + t[0])

            record = data.copy()
            if t[0] in ['DATE', 'TIMESTAMP']:
                # If it's date/timestamp we need to NULL out the other columns and
                # adjust the expected values
                for i in range(1, len(data) + 1):
                    if i in [8, 9]:
                        continue
                    record['c' + str(i)] = None
                if t[0] == 'DATE':
                    record['c9'] = record['c8']
            elif t[0].startswith('VARCHAR') or t[0].startswith('CHAR'):
                # Fix up VARCHAR, the result can't go over the length
                for i in range(1, len(data) + 1):
                    n = int(t[0].split('(')[1].split(')')[0])
                    c = 'c' + str(i)
                    if record[c] is None:
                        continue
                    v = str(record[c])
                    record[c] = v[0:min(len(v), n)]

                    # For CHAR, we need to pad to the max length
                    if t[0].startswith('CHAR'):
                        for _ in range(0, n - len(v)):
                            record[c] += ' '
            elif t[0] == 'DECIMAL(1, 0)':
                # This is not a great way to write these tests
                for i in range(1, len(data) + 1):
                    c = 'c' + str(i)
                    record[c] = None
                record['c1'] = 0
            elif t[0] == 'DECIMAL(10, 1)':
                # This is not a great way to write these tests
                for i in range(1, len(data) + 1):
                    c = 'c' + str(i)
                    record[c] = None
                record['c1'] = 0.0
                record['c4'] = 123.0
            elif t[0] == 'DECIMAL(19, 2)':
                # This is not a great way to write these tests
                for i in range(1, len(data) + 1):
                    c = 'c' + str(i)
                    record[c] = None
                record['c1'] = 0.0
                record['c4'] = 123.0
                record['c5'] = 4.56
            elif t[0] == 'DECIMAL(38, 10)':
                # This is not a great way to write these tests
                for i in range(1, len(data) + 1):
                    c = 'c' + str(i)
                    record[c] = None
                record['c1'] = 0.0
                record['c4'] = 123.00
                record['c5'] = 4.56

            contents = []
            for _ in range(0, 3):
                contents.append(json.dumps(data))

            path = self._write_s3(TEST_DIR_PREFIX, 'type-conversion/data.json',
                                  '\n'.join(contents))
            create = '''
        CREATE EXTERNAL TABLE %s(%s)
        STORED AS JSON
        LOCATION '%s' ''' % \
            (TEST_TABLE, ', '.join(cols), path)
            self._test_path(path, [record, record, record], TEST_DB, TEST_TABLE,
                            sql=create, required_type=t[1],
                            file_format=FILE_FORMAT)

    @unittest.skipIf(common.TEST_LEVEL != "all",
                     "Not very stable, downloads from internet.")
    def test_urls(self):
        for url in TEST_URLS:
            self.__test_url(url[0], url[1])

    @unittest.skipIf(common.TEST_LEVEL != "all", "Skipping at unit test level")
    def test_basic_random_schemas(self):
        self._cleanup(TEST_DIR_PREFIX, 'basic/')
        generator = common.JsonGenerator(
            seed=int(time.time() * 1000))
        self._run_random_tests(generator, None, 'basic')
        # Only cleanup on success, this helps with S3 consistency issues.
        self._cleanup(TEST_DIR_PREFIX, 'basic/')

    @unittest.skipIf(common.TEST_LEVEL != "all", "Skipping at unit test level")
    def test_basic_schema_merge(self):
        self._cleanup(TEST_DIR_PREFIX, 'merge/')
        generator = common.JsonGenerator(
            seed=int(time.time() * 1000),
            min_records=2,
            max_recursion=3,
            max_records=3,
            null_probability=0,
            empty_record_probability=0,
            missing_fields_probability=0,
            generate_variadic_schema=True)
        self._run_random_tests(generator, None, 'merge')
        # Only cleanup on success, this helps with S3 consistency issues.
        self._cleanup(TEST_DIR_PREFIX, 'merge/')

    @unittest.skip("Finding too many bugs")
    def test_medium_random_schemas(self):
        self._cleanup(TEST_DIR_PREFIX, 'medium/')
        generator = common.JsonGenerator(
            seed=int(time.time() * 1000),
            null_probability=.15,
            min_fields=1,
            max_fields=7,
            max_array_len=4,
            max_recursion=6)
        self._run_random_tests(generator, None, 'medium')
        # Only cleanup on success, this helps with S3 consistency issues.
        self._cleanup(TEST_DIR_PREFIX, 'medium/')

    @unittest.skipIf(common.TEST_LEVEL != "all", "Skipping at unit test level")
    def test_long_running(self):
        # Runs the test for a fixed amount of time or until the max failures are
        # hit. This will archive the test files to S3 for async debugging.
        print("\n\n----------------------------------------------------------------")
        generator = common.JsonGenerator(
            seed=int(time.time() * 1000),
            null_probability=.05,
            empty_record_probability=.05,
            missing_fields_probability=.05,
            min_fields=1,
            max_fields=7,
            max_array_len=4,
            max_recursion=6,
            generate_empty_record_all_types=True,
            generate_variadic_schema=True)
        self._test_long_running(generator)

    @unittest.skipIf(common.TEST_LEVEL != "all", "Skipping at unit test level")
    def test_larger_long_running(self):
        # Runs the test for a fixed amount of time or until the max failures are
        # hit. This will archive the test files to S3 for async debugging.
        print("\n\n----------------------------------------------------------------")
        generator = common.JsonGenerator(
            seed=int(time.time() * 1000),
            null_probability=.05,
            empty_record_probability=0.01,
            missing_fields_probability=.05,
            min_fields=3,
            max_fields=7,
            max_array_len=5,
            max_recursion=6,
            min_records=100,
            max_records=200,
            generate_empty_record_all_types=True,
            generate_variadic_schema=True)
        self._test_long_running(generator)

    @unittest.skipIf(common.TEST_LEVEL != "all", "Skipping at unit test level")
    def test_even_larger_long_running(self):
        # Runs the test for a fixed amount of time or until the max failures are
        # hit. This will archive the test files to S3 for async debugging.
        print("\n\n----------------------------------------------------------------")
        generator = common.JsonGenerator(
            types=[[TTypeId.STRING]],
            seed=int(time.time() * 1000),
            null_probability=.05,
            empty_record_probability=0,
            missing_fields_probability=0,
            min_fields=1,
            max_fields=2,
            max_array_len=5,
            max_recursion=1,
            min_records=10000,
            max_records=50000,
            generate_empty_record_all_types=False,
            generate_variadic_schema=False)
        self._test_long_running(generator)

    def _test_long_running(self, generator):
        # Max failures and duration to run
        max_failures = 5
        duration_sec = 30
        test_presto = DEFAULT_ENABLE_PRESTO
        if 'JSON_TEST_MAX_FAILURES' in os.environ:
            max_failures = int(os.environ['JSON_TEST_MAX_FAILURES'])
        if 'JSON_TEST_DURATION_SECS' in os.environ:
            duration_sec = int(os.environ['JSON_TEST_DURATION_SECS'])
        if 'JSON_TEST_PRESTO' in os.environ:
            test_presto = os.environ['JSON_TEST_PRESTO'].lower() == 'true'

        total_tests = 0
        total_success = 0
        total_skips = 0
        failures = []

        # Generate a random directory to store failures
        start = datetime.now()
        directory = start.strftime("%Y-%m-%d-%H-%M-%S")
        print("Generating test data in: s3://%s/%s/%s/" %\
              (TEST_BUCKET, TEST_DIR_PREFIX, directory))
        print("    Test presto: %s" % test_presto)
        end = start + timedelta(seconds=duration_sec)

        while datetime.now() < end:
            total_tests += 1
            generator.new_schema()

            # If testing with presto, we need to generate with a record idx so we can
            # sort the result set for comparison.
            data = generator.generate(generate_record_idx=test_presto)
            filename = '%s/test-case-%s/data.json' % (directory, total_tests)

            output = []
            for record in data:
                output.append(json.dumps(record))
            full_path = self._write_s3(TEST_DIR_PREFIX, filename, '\n'.join(output))

            sql, skip = self._test_path(
                full_path, data, TEST_DB, TEST_TABLE, allow_missing=True, batch_mode=True,
                test_presto=test_presto, sort_by_record_idx=test_presto,
                file_format=FILE_FORMAT)
            if not sql:
                total_success += 1
                continue
            if skip:
                total_skips += 1
                continue
            failures.append((full_path, sql))

            if len(failures) >= max_failures:
                break

        print("\n----------------------------------------------------------------")
        print("Total tests duration: %s" % (datetime.now() - start))
        print("Total tests runs: %s" % total_tests)
        print("Total successful: %s" % total_success)
        print("Total skips: %s" % total_skips)
        print("Total failures: %s" % len(failures))

        if not failures:
            return

        print("Tests which failed.\n")
        for (path, sql) in failures:
            print("----------------------------------------------------------------")
            print("Path: %s" % path)
            print("SQL: %s\n" % sql)

        self.fail("Found failures")



# Copyright 2020 Okera Inc. All Rights Reserved.
#
# Some integration tests for hive in PyOkera
#
# pylint: disable=logging-not-lazy
# pylint: disable=no-self-use
# pylint: disable=no-else-return
import os
import unittest
import time

from pyhive import hive
from okera.tests import pycerebro_test_common as common
from okera.tests import _mr_util as mr_util
from okera._util import get_logger_and_init_null
from thrift.transport.TTransport import TTransportException

DEFAULT_HIVE_HOST = os.environ['ODAS_TEST_HOST']
DEFAULT_HIVE_PORT = int(os.environ['ODAS_TEST_PORT_HIVESERVER2'])
IS_AUTHORIZE_QUERY = common.get_bool_env_var('OKERA_ENABLE_AUTHORIZE_QUERY', False)

_log = get_logger_and_init_null(__name__)

blacklisted_dbs = [
    'delta_db',
    # Snowflake tables can cause high resource utilization on snowflake
    'tpcds001tb_snowflake',
    'sf_tpcds_1gb',
    'tpch_sf1',
    'tpch_sf5',
    'tpch_test_snowflake',
]

# blacklist tables that we do not support scanning via hive. revisit.
blacklisted_tbls = [
    'bad_metadata_db.broken_view',
    'chase.redacted_ledger_balance_view',
    'chase.cte_view',
    'customer.zd608_fiducia_risk_score_ato_tb',
    'demo_test.transactions_anonymize_inactive_users_external_view',
    'okera_system.configs',
    'okera_system.events_by_user_dataset_tags',
    'okera_system.fs_events_by_client_ip_tagged',
    'okera_system.fs_events_by_hour_tags',
    'okera_system.fs_events_by_user_dataset_tags',
    'okera_system.steward_fs_tagged_audit_logs',
    'okera_system.steward_fs_tagged_audit_logs_preview',
    'partition_test.timestamp_part_encoded',
    'partition_test.timestamp_part_test',
    'partition_test.weird_partition2',
    'partition_test.weird_partition3',
    'partition_test.weird_partition4',
    'partition_test.weird_partition5',
    'partition_test.weird_partition6',
    'partition_test.weird_partition7',
    'partition_test.weird_partition8',
    'rs.partitioned_alltypes',
    'special_chars.test_table',

    # Non standard partition files. Its bizzare that the hudi_partitioned table failed!
    'rs_complex_parquet.hudi_partitioned',
    'rs_complex_parquet.hudi_nonpartitioned',
    'rs_complex_parquet.hudi_as_parquet',
]

# Authorize query has some additional requirements
authorize_query_blacklisted_tbls = [
    # The complex view contains constructs Hive cannot understand.
    'authdb.struct_t_full_view_on_sel_complex_view',
    'authdb.struct_t_sel_view_on_sel_complex_view',
    'authdb.struct_t_view_on_view',
    'customer.market_decide_offer_decision_v3_card_non_npi_vw',

    # Hive can't read spark bucketed tables
    'buckets_test.spark_bucketed_test_view',
] + mr_util.authorize_query_unsupported_datasets

def skip_table(tbl_name):
    if tbl_name in blacklisted_tbls:
        return True
    if IS_AUTHORIZE_QUERY and tbl_name in authorize_query_blacklisted_tbls:
        return True
    return False

class HiveScanTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """ Initializes one time state that is shared across test cases. This is used
            to speed up the tests. State that can be shared across (but still stable)
            should be here instead of __cleanup()."""
        super(HiveScanTest, cls).setUpClass()
        # retry for 30 seconds.
        maxRetryCnt = 6
        currRetryCnt = 0
        while (True):
            try:
                cls._conn = hive.Connection(host=DEFAULT_HIVE_HOST,
                                            port=DEFAULT_HIVE_PORT, username="root")
                cls._cursor = cls._conn.cursor()
                break
            except TTransportException as ex:
                if (currRetryCnt >= maxRetryCnt):
                    print("Hive connection failed. All retries exhausted, aborting.")
                    raise ex
                currRetryCnt += 1
                print("Hive connection failed. Retrying after 5 seconds.")
                time.sleep(5)


    @classmethod
    def tearDownClass(cls):
        cls._cursor.close()
        cls._conn.close()

    def __describe(self, tbl):
        self._cursor.execute('DESCRIBE FORMATTED %s' % tbl)
        return '\n'.join([str(s) for s in self._cursor.fetchall()])

    def __explain(self, tbl):
        self._cursor.execute('EXPLAIN SELECT * FROM %s' % tbl)
        return '\n'.join([str(s) for s in self._cursor.fetchall()])

    def __query_first_row(self, sql):
        self._cursor.execute(sql)
        return self._cursor.fetchall()[0]

    @unittest.skipIf(not IS_AUTHORIZE_QUERY, "Skipping, authorize query is disabled.")
    def test_show_okera_udfs(self):
        self._cursor.execute("SHOW FUNCTIONS like 'okera_udfs.*'")
        rows = self._cursor.fetchall()
        udfs = []
        for row in rows:
            udfs.append(row[0])
        print(udfs)
        self.assertTrue('okera_udfs.mask' in udfs)
        self.assertTrue('okera_udfs.mask_ccn' in udfs)
        self.assertTrue('okera_udfs.null' in udfs)
        self.assertTrue('okera_udfs.sha2' in udfs)
        self.assertTrue('okera_udfs.tokenize' in udfs)
        self.assertTrue('okera_udfs.zero' in udfs)

    @unittest.skipIf(not IS_AUTHORIZE_QUERY, "Skipping, authorize query is disabled.")
    def test_okera_udfs(self):
        self.assertEqual(
            'xxxx',
            self.__query_first_row("select okera_udfs.mask('abcd')")[0])
        self.assertEqual(
            'XXXX-efgh',
            self.__query_first_row("select okera_udfs.mask_ccn('abcd-efgh')")[0])
        self.assertEqual(
            None,
            self.__query_first_row("select okera_udfs.`null`('abcd')")[0])
        self.assertEqual(
            267444780464769297,
            self.__query_first_row("select okera_udfs.sha2('abcd')")[0])
        self.assertTrue(
            self.__query_first_row("select okera_udfs.tokenize('abcd')")[0] != 'abcd')
        self.assertEqual('', self.__query_first_row("select okera_udfs.zero('abcd')")[0])

    def test_show_databases(self):
        self._cursor.execute('SHOW DATABASES')
        rows = self._cursor.fetchall()
        dbFound = False
        for row in rows:
            if 'okera_sample' in row:
                dbFound = True
                break
        self.assertTrue(dbFound)

    def test_show_tables(self):
        self._cursor.execute('SHOW DATABASES')
        dbs = self._cursor.fetchall()
        for db in dbs:
            if not db[0].startswith('_'):
                self._cursor.execute('SHOW TABLES IN ' + db[0])

    def test_select_singlerow(self):
        tables_scanned = 0
        self._cursor.execute('SHOW DATABASES')
        dbs = []
        if common.test_level_lt('checkin'):
            # At lower levels, test only some dbs
            dbs = ['authdb', 'jdbc_test_mysql', 'okera_sample', 'rs', 'rs_complex']
        else:
            for db in self._cursor.fetchall():
                db = db[0]
                if db.startswith('_') or db in blacklisted_dbs:
                    continue
                dbs.append(db)

        print("Hive::test_select_singlerow DBs: " + str(dbs))
        for db in dbs:
            self._cursor.execute('SHOW TABLES IN ' + db)
            tbls = self._cursor.fetchall()
            for tbl in tbls:
                full_tbl_name = db + '.' + tbl[0]
                if tbl[0].startswith('_') or skip_table(full_tbl_name):
                    continue
                tables_scanned += 1
                sql = 'SELECT * FROM ' + full_tbl_name + ' LIMIT 1'
                print("Hive::test_select_singlerow SQL: " + sql)
                self._cursor.execute(sql)
        _log.info("Total tables scanned: %d" % tables_scanned)

    @unittest.skip("Used for testing to run a single table easily.")
    def test_select_singlerow_dev(self):
        tbl = 'demo_test.transactions_active_users'
        sql = 'SELECT * FROM ' + tbl + ' LIMIT 1'
        self._cursor.execute(sql)

    @unittest.skipIf(not IS_AUTHORIZE_QUERY, "Skipping, authorize query is disabled.")
    def test_explain(self):
        whoami = self.__describe('okera_sample.whoami')
        self.assertTrue("SELECT 'root' as `user`" in whoami)
        whoami = self.__explain('okera_sample.whoami')
        self.assertTrue("alias: _dummy_table" in whoami)
        self.assertTrue("expressions: 'root'" in whoami)

        users_ccn_masked = self.__describe('okera_sample.users_ccn_masked')
        print(users_ccn_masked)
        self.assertTrue("okera_udfs.mask_ccn(`ccn`) FROM" in users_ccn_masked)

        alltypes_s3 = self.__describe('rs.alltypes_s3')
        self.assertTrue("s3a://cerebrodata-test/alltypes" in alltypes_s3)
        self.assertTrue("org.apache.hadoop.mapred.TextInputFormat" in alltypes_s3)
        alltypes_s3 = self.__explain('rs.alltypes_s3')
        self.assertTrue("alias: alltypes_s3" in alltypes_s3)
        self.assertTrue("expressions: bool_col (type: boolean)" in alltypes_s3)
